#!/usr/bin/perl
###############################################################################
#                                                                             #
# halfStats.pl - Converts Half-Life logs to HTML statistics pages             #
#              - Copyright Zach Karpinski                                     #
#              - Distributed under the GPL terms - see the docs for info      #
#              - http://www.halfstats.com                                     #
#                                                                             #
###############################################################################
# CVS $Id: halfStats.pl,v 1.18 2001/03/21 02:52:00 smccrory Exp $
###############################################################################
#
#    halfStats
#    Copyright (C) 2000 Zach Karpinski <zkarpinski@yahoo.com>
#    All Rights Reserved
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program; if not, write to the Free Software
#    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
###############################################################################

use strict;             # Prevents the usage of variable without declaration
use Player;             # Grab methods which construct the data structures
use Data::Dumper;       # Used to save and restore pre-parsed data structures
use Time::Local;        # Gives us local time in epoch seconds

######################## GLOBAL CONSTANTS #####################################

# See sub setConstants for most of these...

###################### GLOBAL DEVELOPMENT CONSTANTS ###########################

# Development constants - DON'T mess with these please
my $SCRIPT_NAME         = "halfStats";
my $VERSION             = "1.0 Beta 7";
my $DATE                = "March 20, 2001";

# Many thanks to Scott and Brian.
my $AUTHOR_NAME         = "Zach Karpinski & the halfStats development team";
my $AUTHOR_EMAIL        = "zkarpinski\@yahoo.com";
my $HOME_PAGE           = "http://www.halfstats.com";

############################# GLOBAL VARIABLES ################################

# This stuff is global cause we use it a lot.
my %config;                 # Hash to store all the config vars
my %awards;                 # Hash to store awards

my @logs;                   # Array of log files to process
my @ignored;                # Array of players to ignore
my @players;                # Array of all VALID player keys pointing to data

my $cfgFile;                # Filename of config file
my $mode;                   # How we're storing players (WON, Name, IP)
my $stats;                  # Container for pointer to stats objects
my $sTime;                  # halfStats start time
my $VAR1;                   # Data dumper global variable

# Global Half-Life logging standard variables
my ($team,$player,$event1,$noun1,$event2,$noun2,$properties,%names); 

################################## MAIN #######################################

printBanner();              # Print the program banner
$sTime = time();            # Get current time
$cfgFile = shift;           # Pop the config file off the parameter list

if($cfgFile)                # If the user supplied arguments,
{
   my $dstart;              # Used to hold halfStats' local start time

   # Parse the config file (supplied in $ARGV[0]) and get a hash ref
   # for the stats datastructure
   $dstart = time();

   print "Parsing config file...\n";
   parseConfig();

   print "Validating config values...\n";
   my $validConfig = checkConfig();
   if ( $validConfig ne "OK")
   {
      print "\tConfig value error: " . $validConfig . "\n";
      print "\tcheck your config file and try again.\n";
      die;
   }
   
   print "Initializing data structure...\n";
   $stats = init();
   print "\t-> Finished in " . calcTime($dstart) . " seconds\n";

   if(getArgs("reset"))
   {
       $dstart = time();
       print "Resetting stored data files...\n";
       resetLogs();
       print "\t-> Finished in " . calcTime($dstart) . " seconds\n";
   }
   else
   {
       # This loads the saved Data::Dump of our \%stats hash
       $dstart = time();
       print "Loading stored data...\n";
       loadStoredData();
       print "\t-> Finished in " . calcTime($dstart) . " seconds\n";
   }

   # Open and process the logs created by the HL server since last run
   $dstart = time();
   print "Reading new logs...\n";
   readNewLogs();
   print "\t-> Finished in " . calcTime($dstart) . " seconds\n";

   # Pull the keys into our @players array to make manipulating them possible
   @players = keys %{$stats->{'players'}};

   # Get list of player names to ignore
   $dstart = time();
   print "Finding out who to ignore...\n";
   loadIgnoredPlayers();
   print "\t-> Finished in " . calcTime($dstart) . " seconds\n";

   # Weed out the deadwood
   $dstart = time();
   print "Weeding out players...\n";
   weedOutPlayers();
   print "\t-> Finished in " . calcTime($dstart) . " seconds\n";

   # Run the routines to create the stats
   $dstart = time();
   print "Running html creation routines...\n";
   print "(We'll put output in " . $config{"abs_html_dir"} . ")\n";
   createStatPages();
   print "\t-> Finished in " . calcTime($dstart) . " seconds\n";

   if(getArgs("archive"))
   {
       $dstart = time();
       print "Archiving server logs...\n";
       archiveLogs();
       print "\t-> Finished in " . calcTime($dstart) . " seconds\n";
   }

   if(getArgs("clean"))
   {
       $dstart = time();
       print "Deleting server logs...\n";
       cleanLogs();
       print "\t-> Finished in " . calcTime($dstart) . " seconds\n";
   }

   # Save the stats data
   if(!getArgs("nosave"))
   {
       $dstart = time();
       print "Writing data files...\n";
       writeData();
       print "\t-> Finished in " . calcTime($dstart) . " seconds\n";
   }

   print "** $SCRIPT_NAME completed in " . calcTime($sTime) . " seconds **\n";
   exit 0;                  # Exit with a successful return code
}
else
{
   printUsage();            # Inform them of proper usage.
   exit 1;                  # Exit with an error return code
}

############################### SUBROUTINES ###################################

#------------------------------------------------------------------------------
# Print the halfStats program banner
sub printBanner
{
    print "\n$SCRIPT_NAME $VERSION, $DATE by $AUTHOR_NAME\n";
    print "Visit $HOME_PAGE for more information\n\n";
}

#------------------------------------------------------------------------------
# Print halfStats argument usage for the user
sub printUsage
{
    print "\nNo arguments specified!\n";
    print "usage: $SCRIPT_NAME configfile <option1> <option2> <..>\n\n";
    print "       Options:  archive    - moves logs into config-defined dir after processing\n";
    print "                 reset      - Resets (deletes) the *stored* player stats,\n";
    print "                              causing them to be rebuilt fresh from the logs\n";
    print "                 nosave     - prevents saving of stats data\n";
    print "                 clean      - cleans out OLD game log files (CAUTION!)\n";
    print "                 clean wipe - wipes out ALL game log files (MAJOR CAUTION!!)\n";
    print "\n";
    print " Example 1: $SCRIPT_NAME halfStats.cfg\n";
    print "    - process only NEW logs since last run and save our place in data files\n";
    print " Example 2: $SCRIPT_NAME halfStats.cfg reset nosave\n";
    print "    - process ALL logs without loading old data or saving our place\n\n";
}

#------------------------------------------------------------------------------
# Search the command-line parameters for a specific argument and
# return true if there is a match
sub getArgs
{
    my $searchKey = shift;
    my $arg = "";
    my $result = 0;
    
    foreach $arg (@ARGV)
    {
        if ($arg eq $searchKey) { $result = 1 ; last; }
    }
    return $result;
}

#------------------------------------------------------------------------------
# Grab all the user preferences into the global %config hash
sub parseConfig
{
    my ($var, $val);
    if(!$cfgFile) { $cfgFile = "halfStats.cfg"; }
    open(CONF, "<$cfgFile") || die "Cannot find config file: $!\n";
    while (<CONF>) {
        chomp;                      # no newline
        s/#.*//;                    # no comments
        s/^\s+//;                   # no leading white
        s/\s+$//;                   # no trailing white
        next unless length;         # anything left?
        ($var, $val) = split(/\s*=\s*/, $_, 2);
        $config{$var} = $val;       # load the untainted value into our hash of vars
    }
    close(CONF);
}

#------------------------------------------------------------------------------
# Check to make sure all the required user preferences are present and valid
sub checkConfig
{
    if ($config{"os"} ne "unix" && $config{"os"} ne "windows")
    {
      return ("'os' value must be unix or windows");
    }

    if (!$config{"logdir"})
    {
      return ("'logdir' is missing or invalid");
    }

    if ($config{"game_type"} ne "cstrike" && $config{"game_type"} ne "tfc" && $config{"game_type"} ne "flf" && $config{"game_type"} ne "dm")
    {
      return ("'game_type' value missing or not supported");
    }

    if (!$config{"abs_html_dir"})
    {
      return ("'logdir' is missing or invalid");
    }

    if (!$config{"archive_dir"})
    {
      return ("'archive_dir' is missing or invalid");
    }

    if (!$config{"abs_img_dir"})
    {
      return ("'abs_img_dir' is missing or invalid");
    }

    if (!$config{"rel_img_dir"})
    {
      return ("'rel_img_dir' is missing or invalid");
    }

    if (!$config{"server_name"})
    {
      return ("'server_name' is missing or invalid");
    }
    
    if (!$config{"css_color"})
    {
      return ("'css_color' is missing or invalid");
    }

    if (!$config{"minimum_kills"} || $config{"minimum_kills"} < 0)
    {
      return ("'minimum_kills' is missing or invalid");
    }

    if (!$config{"minimum_skill"} || $config{"minimum_skill"} < 0)
    {
      return ("'minimum_skill' is missing or invalid");
    }

    if (!$config{"rank_max"} || $config{"rank_max"} < 0)
    {
      return ("'rank_max' is missing or invalid");
    }

    if (!$config{"minimum_visit_frequency"} || $config{"minimum_visit_frequency"} < 0)
    {
      return ("'minimum_visit_frequency' is missing or invalid");
    }
    
    if (!$config{"minimum_minutes_played"} || $config{"minimum_minutes_played"} < 0)
    {
      return ("'minimum_minutes_played' is missing or invalid");
    }
    
    if (!$config{"ignored_players_file"})
    {
      return ("'ignored_players_file' is missing or invalid");
    }
    
    if (!$config{"admin_name"})
    {
      return ("'admin_name' is missing or invalid");
    }

    if (!$config{"admin_email"})
    {
      return ("'admin_email' is missing or invalid");
    }
    
    if ($config{"store_method"} ne "wonid" && $config{"store_method"} ne "ip" && $config{"store_method"} ne "name" )
    {
      return ("'store_method' is missing or invalid");
    }
    
    if ($config{"show_ip"} ne "true" && $config{"show_ip"} ne "false" )
    {
      return ("'show_ip' is missing or invalid");
    }

    if (!$config{"sleep_interval"} || $config{"sleep_interval"} < 0)
    {
      return ("'sleep_interval' is missing or invalid");
    }

    if (!$config{"sleep_time"} || $config{"sleep_time"} < 0)
    {
      return ("'sleep_time' is missing or invalid");
    }

    if ($config{"hlmaps"} ne "true" && $config{"hlmaps"} ne "false" )
    {
      return ("'hlmaps' is missing or invalid");
    }

    if (!$config{"hlmaps_url"})
    {
      return ("'hlmaps_url' is missing or invalid");
    }

    return ("OK");
}

#------------------------------------------------------------------------------
# Deletes ALL HL server log files AND player stats (data files) if specified
sub resetLogs
{
    if(getArgs("wipe"))
    {
        cleanLogs();
    }
    clearPlayersData();
    undef $stats;
}

#------------------------------------------------------------------------------
# Archives ALL old log files
sub archiveLogs
{
    my $log;
    my(@files,$file);
    my $unlinkcount;

    print "Now archiving logs...\n";

    if (!(-e "$config{'archive_dir'}"))
    {
	mkdir ($config{'archive_dir'}, 0770);
    }

    foreach $log (keys %{$stats->{'old_logs'}})
    {
        print "\t\tArchiving $log\n";
        rename("$config{'logdir'}/$log","$config{'archive_dir'}/$log") or warn "Could not rename $config{'logdir'}/$log!";
    }
    print "Stored logs now archived..\n";
}

#------------------------------------------------------------------------------
# Deletes ALL old log files
sub cleanLogs
{
    my $log;
    my(@files,$file);
    my $unlinkcount;

    print "\tNow deleting logs...\n";

    if(getArgs("wipe"))
    {
        opendir(DIR, "$config{'logdir'}") || die "can't open $config{'logdir'} dir for log deletion!\n";
        @files = grep { /.log$/ }  readdir(*DIR);
        close(DIR);
        
        unless (($unlinkcount = unlink(@files)) == @files)
        {
            warn "Could only delete $unlinkcount of " . (@files) . " log files!";
        }
        print "ALL log files wiped out..\n\n";
    }
    else
    {
        foreach $log (keys %{$stats->{'old_logs'}})
        {
            print "\t\tDeleting $log\n";
            unlink("$config{'logdir'}/$log") or warn "Could not delete $config{'logdir'}/$log!";
        }
        print "Stored logs now purged..\n";
    }
}

#------------------------------------------------------------------------------
# Clear out stored player statistics by deleting all their data files
sub clearPlayersData
{
    my(@files,$file);
    my $unlinkcount;

    # Create the data/players/$cfgFile directory if it doesn't exist
    if(!(-e "data/players/$cfgFile"))
    {
        if($config{'os'} eq "unix")
        {
            mkdir ("data", 0770);
            mkdir ("data/players", 0770);
            mkdir ("data/players/$cfgFile", 0770);
        }
        else
        {
            mkdir ("data", 0770);
            mkdir ("data\\players", 0770);
            mkdir ("data\\players\\$cfgFile", 0770);
        }
    }
    
    if($config{'os'} eq "unix")
    {
        unlink("data/$cfgFile-live-logData.txt");
        unlink("data/$cfgFile-server-logData.txt");
        unlink("data/$cfgFile-old_logs-logData.txt");

        opendir(DIR, "data/players/$cfgFile") || die "Directory data/players/$cfgFile does not exist - please create it!\n";
        @files = grep { /.dat$/ }  readdir(*DIR);
        close(DIR);
        
        foreach $file (@files)
        {
            unlink("data/players/$cfgFile/$file") or warn "Could not delete data/players/$cfgFile/$file";
        }
    }
    else
    {
        unlink("data\\$cfgFile-live-logData.txt");
        unlink("data\\$cfgFile-server-logData.txt");
        unlink("data\\$cfgFile-old_logs-logData.txt");

        opendir(DIR, "data\\players\\$cfgFile") || die "Directory data\\players\\$cfgFile does not exist - please create it!\n";
        @files = grep { /.dat$/ }  readdir(*DIR);
        close(DIR);
        
        foreach $file (@files)
        {
            unlink("data\\players\\$cfgFile\\$file") or warn "Could not delete data\\players\\$cfgFile\\$file";
        }
    }
    print "\tDeleted player stats data files\n";
}

#------------------------------------------------------------------------------
# Initialize Data::Dumper and get the configuration preferences
sub init
{
    # Since we are using Data::Dumper we're going to start with a hash reference
    my (%start,@array,$line);

    local $Data::Dumper::Purity = 1;

    $mode = $ARGV[0];
    if(!$config{'rank_max'}) { $config{'rank_max'} = 100; }
    return(\%start);
}

#------------------------------------------------------------------------------
# Use Data::Dumper to load in all our saved structures from disk
sub loadStoredData
{
    
    my (@array, $string, $line,$file,$plr,$count);

    if(-e "data/$cfgFile-live-logData.txt"
       && -e "data/$cfgFile-server-logData.txt"
       && -e "data/$cfgFile-old_logs-logData.txt"
       )
    {
        $count = 0;
    
        # Create the data/players directory if it doesn't exist
        if(!(-e "data/players/$cfgFile"))
        {
            if($config{'os'} eq "unix")
            {
                mkdir ("data", 0770);
                mkdir ("data/players", 0770);
                mkdir ("data/players/$cfgFile", 0770);
            }
            else
            {
                mkdir ("data", 0770);
                mkdir ("data\\players", 0770);
                mkdir ("data\\players\\$cfgFile", 0770);
            }
        }
    
        # Load in each one of the player data files
        opendir(DIR, "data/players/$cfgFile") || die "data/players/$cfgFile does not exist\n";
        my @files = grep { /.dat$/ }  readdir(*DIR);
        close(DIR);
    
        foreach $file(@files)
        {
            open(FILE, "<data/players/$cfgFile/$file");
            @array = <FILE>;
            close(FILE);
    
            $string = makeString(\@array);
    
            my($plr,$ext) = split(/\./, $file);
            
            $stats->{'players'}{$plr} = eval $string;
            
            undef $string;undef @array;
        }
    
        # Pull in the live log information
        open(FILE, "<data/$cfgFile-live-logData.txt");
        @array = <FILE>;
        close(FILE);

        $string = makeString(\@array);
        $stats->{'live'} = eval $string;    
        undef $string;undef @array;
        
        # Pull in the stored server information
        open(FILE, "<data/$cfgFile-server-logData.txt");
        @array = <FILE>;
        close(FILE);

        $string = makeString(\@array);
        $stats->{'server'} = eval $string;
        undef $string;undef @array;
        
        # Pull in the old logs information
        open(FILE, "<data/$cfgFile-old_logs-logData.txt");
        @array = <FILE>;
        close(FILE);

        $string = makeString(\@array);
        $stats->{'old_logs'} = eval $string;
        undef $string;undef @array;
        
        # Now weed out obsolete old_logs
        weedOutMissingLogs();
    }
}

#------------------------------------------------------------------------------
# Take all elements of an array and create a single concatenated string of them
sub makeString
{
    my ($string, $array, $line);
    $array = shift;
    $string .= "";
    foreach $line (@$array)
    {
        $string .= $line;
    }
    return($string);
}

#------------------------------------------------------------------------------
# Get rid of any references in the old_logs structure if the log file doesn't
#     presently exist on the system.  This is to prevent this array from
#     growing out of control.
sub weedOutMissingLogs
{
    my $logfile = '';
    
    foreach $logfile (keys %{$stats->{'old_logs'}})
    {
        unless (-e "$config{'logdir'}/$logfile")
        {
            print "\tRemoving reference to non-existent log $logfile\n";
            delete $stats->{'old_logs'}{$logfile};
        }
    }
}

#------------------------------------------------------------------------------
# Sort by file's modification date
sub byFileModDate
{
    my ($dev_1,$ino_1,$mode_1,$nlink_1,$uid_1,$gid_1,$rdev_1,$size_1,$atime_1,
        $mtime_1,$ctime_1,$blksize_1,$blocks_1)
            =stat("$config{'logdir'}/$a");
    
    my ($dev_2,$ino_2,$mode_2,$nlink_2,$uid_2,$gid_2,$rdev_2,$size_2,$atime_2,
        $mtime_2,$ctime_2,$blksize_2,$blocks_2)
            =stat("$config{'logdir'}/$b");

    if ($mtime_1 < $mtime_2) {
        return -1;
    } elsif ($mtime_1 == $mtime_2) {
        return 0;
    } elsif ($mtime_1 > $mtime_2) {
        return 1;
    }
}

#------------------------------------------------------------------------------
# This is the main log processing subroutine.
sub readNewLogs
{
    my ($line,%tally,@array,$count,$logfile,$mon,$day,$year,$hour,$min,$sec,$record);

    $count = 0;

    opendir(DIR, $config{'logdir'}) || die "can't opendir $! [$config{'logdir'}]\n";
    my @files = grep { /L\d+\.log$/ } readdir(*DIR);
    close(DIR);

    foreach $logfile (sort byFileModDate @files)
    {

        my ($dev,$ino,$mode,$nlink,$uid,$gid,$rdev,$size,$atime,$mtime,$ctime,$blksize,$blocks)
            =stat("$config{'logdir'}/$logfile");
    
        if($logfile =~ ".log" && !$stats->{'old_logs'}{$logfile})
        {
            $stats->{'server'}{'general'}{'log_bytes'} += $size;
    
            open(FILE, "<$config{'logdir'}/$logfile");
            @logs = grep {!/<-1>/} <FILE>;
            close(FILE);
    
            # Only process the log if it is complete!
            if($logs[$#logs] && $logs[$#logs] =~ "Log file closed")
            {
                my ($inCvar) = 0;   # Reset the CVAR flag
                print "\tParsing  $logfile\n";
                
                foreach $line (@logs)
                {
                    chomp($line);
                    
                    # Ignore server CVAR settings group
                    if($line =~ 'server cvars start')   { $inCvar = 1; }
                    if($line =~ 'server cvars end')     { $inCvar = 0; }
                    next unless !$inCvar;                       # Ignore server CVAR settings group

                    # Ignore a variety of other lines for efficiency's sake
                    
                    my $time = getSecondsFromLogFormat("$line");
                    
                    $line =~ /^L (\d\d)\/(\d\d)\/(\d\d\d\d) - (\d\d):(\d\d):(\d\d): (.+)/;
                    $mon = $1; $day = $2; $year = $3; $hour = $4; $min = $5; $sec = $6; $record = $7;

                    if($record)
                    {
                        # Strip off excess returns found in Windows logs (Thanks Ryan!)
                        $record =~ s/\r//g;
                        
                        parseline($record);

                        if($player) 
                        {
                          checkPlayer();
                        }

                        #print "Just parsed:\n\tteam:[$team]\n\tplayer:[$player]\n\te1:[$event1]\n\tn1:[$noun1]\n\te2:[$event2]\n\tn2:[$noun2]\n\tprop:[$properties]\n\n";

                        # Now begin the heavy-duty parsing...
                        if($record =~ '\>" killed "' && $record =~ '" with ')
                        {
                            #"*crimrOw*-[shOck]-<3>" killed "daf<2>" with tau_cannon
                            
                            #my($killer,$kNum,$victim,$vNum,$weapon) = ($1,$2,$3,$4,$5);
                           my($killer,$kNum,$kId,$kTeam) = getPlayerInfo($player);
                           my($victim,$vNum,$vId,$vTeam) = getPlayerInfo($noun1);

                           if($killer eq "world")
                           {
                             ($killer,$kNum,$kId,$kTeam) =  ($victim,$vNum,$vId,$vTeam);
                           }

                           my($useKid,$useVid) = ($names{cleanName($killer)},$names{cleanName($victim)});

                           if(!$useKid || !$useVid || !$stats->{'players'}{$useKid} || !$stats->{'players'}{$useVid})
                           {
                              next;
                           }

                           my($weapon) = $noun2;

                           if(!$killer || !$kNum || !$kId || !$victim || !$vNum || !$vId || !$weapon ) { next; }

                            if($kId && $vId && length($killer) >= 1 && length($victim) >= 1)
                            {
                                $stats->{'server'}{'weapons'}{$weapon} += 1;
                                $stats->{'players'}{$useKid}->kill(\%{$stats->{'players'}{$useVid}},$weapon,$stats->{'live'}{'currentMap'});
                            }
                            else
                            {
                                print "\t\tNotice:\n\tInvalid record: $record\n";
                            }
                            if($stats->{'live'}{'currentMap'})
                            {
                                $stats->{'server'}{'maps'}{$stats->{'live'}{'currentMap'}}{'kills'} += 1;
                            }
                            $stats->{'server'}{'general'}{'kills'} +=1;
                        }
                        elsif($record =~ " connected, address ")
                        {
                            #"daf<15><WON:20341049>" connected, address "192.168.0.10:27005"
                            
                            #$record =~ /^"(.+?)\<(\d+?)\>\<WON:(\d+?)\>" connected, address "(.+?)\:(\d+?)"$/;
                            
                            # Make sure we have all the elements that we need
                            #if (!$1 || !$2 || !$3 || !$4 || !$5) { next; }
                           
                            #my($name,$pNum,$wonId,$ip,$port) = ($1,$2,$3,$4,$5);
                            my($name,$pNum,$wonId,$team) = getPlayerInfo($player);
                            my($ip,$port) = split(":", $noun1);
                            
                            # print "\t\t*** $name connected!!\n\n";

                            if(!$name || !$pNum || !$wonId || !$ip || !$port) { next; }

                            my $cleanName = cleanName($name);
                            
                            my $wonIdToSave = $wonId;
                            
                            if($config{'store_method'} eq "ip")
                            {
                                $stats->{'live'}{$pNum} = $ip;
                                $names{$cleanName} = $ip;
                                $wonId = $ip;
                            }
                            elsif($config{'store_method'} eq "name")
                            {
                                $stats->{'live'}{$pNum} = $cleanName;
                                $names{$cleanName} = $cleanName;
                                $wonId = $cleanName;
                            }
                            
                            $stats->{'live'}{$pNum} = $wonId;
                            $names{$cleanName} = $wonId;
                            
                            if(!$pNum) { print "\t\tNotice: No player number\n"; }
                            if(!$wonId) { print "\t\tNotice: No Won ID\n"; }
                            
                            if($wonId && !$stats->{'players'}{$wonId}{'exists'})
                            {
                                $stats->{'players'}{$wonId} = new Player();
                                if($config{'game_type'} ne "cstrike")
                                {
                                    undef $stats->{'players'}{$wonId}{'cstrike'};
                                }
                                   
                                $stats->{'players'}{$wonId}->setWonId($wonId);
                                $stats->{'players'}{$wonId}->saveWonId($wonIdToSave);
                                $stats->{'server'}{'general'}{'unique'} += 1;
                            }
                            
                            if($stats->{'players'}{$stats->{'live'}{$pNum}}->{'start_time'})
                            {
                                $stats->{'players'}{$stats->{'live'}{$pNum}}->addToTime($time);
                            }
                            
                            $stats->{'players'}{$stats->{'live'}{$pNum}}->recordStartTime($time);
                            $stats->{'players'}{$stats->{'live'}{$pNum}}->setName($name);
                            $stats->{'players'}{$stats->{'live'}{$pNum}}->connect($ip,$time);
                            $stats->{'server'}{'general'}{'connects'} += 1;                     
                        }
                        elsif($record =~ ">\" has entered the game")
                        {
                            #"Captain Salami<15>" has entered the game
                            # -OR-
                            
                            # We take care of most of this upon initial connection
                            # Not much to do?
                            #$record =~ /"(.+?)\<(\d+?)\>" has entered the game/;

                            my ($name, $pNum, $wonId, $team) = getPlayerInfo($player);
                            
                            # Make sure we have all our elements
                            if (!$name || !$pNum) { next; }
                            
                            my $cleanName = cleanName($name);
                            
                            if($config{'store_method'} eq "name")
                            {
                                $stats->{'live'}{$pNum} = $cleanName;
                                $names{$cleanName} = $cleanName;
                                my $wonId = $cleanName;
                                
                                if($wonId && !$stats->{'players'}{$wonId}{'exists'})
                                {
                                    $stats->{'players'}{$wonId} = new Player();
                                    if($config{'game_type'} eq "dm" ||
                                       $config{'game_type'} eq "flf")
                                    {
                                        undef $stats->{'players'}{$wonId}{'cstrike'};
                                    }
                
                                    $stats->{'players'}{$wonId}->setWonId($wonId);
                                }
                                else
                                {
                                   next;
                                }
                                $stats->{'players'}{$wonId}->setName($name);
                                $stats->{'players'}{$wonId}->connect("",$time);
                                $stats->{'server'}{'general'}{'connects'} += 1;
                            }
                            else
                            {
                                $names{$cleanName} = $stats->{'live'}{$pNum};
                            }
                            
                            if($stats->{'live'}{$pNum} && length($stats->{'live'}{$pNum}) >= 1) 
                                {
                                if($stats->{'players'}{$stats->{'live'}{$pNum}}->{'start_time'})
                                {
                                    $stats->{'players'}{$stats->{'live'}{$pNum}}->addToTime($time);
                                }
                                
                                $stats->{'players'}{$stats->{'live'}{$pNum}}->recordStartTime($time);
                            }
                            else
                            {
                                print "\t\tNotice:\n\tInvalid record: $record\n\t no wonid..\n";
                            }
                        }
                        elsif($record =~ ">\" disconnected")
                        {
                            #"edward penishands<15>" disconnected
                            #$record =~ /^"(.+?)\<(\d+?)\>" disconnected/;
                            
                            # Make sure we have all the elements that we need
                            my ($name, $pNum, $wonId, $team) = getPlayerInfo($player);
                            
                            # Make sure we have all our elements
                            if (!$name || !$pNum) { next; }

                            my $cleanName = cleanName($name);
                            
                            if($stats->{'live'}{$pNum} && length($stats->{'live'}{$pNum}) >= 1)
                            {
                                # Add connection time to player's stats
                                $stats->{'players'}{$stats->{'live'}{$pNum}}->addToTime($time);
                                
                                # If possible, update their last-time-played timestamp too
                                # This helps account for the occational malformed log file
                                if( $config{'store_method'} eq "name" && $stats->{'players'}{$cleanName} )
                                {
                                    $stats->{'players'}{$cleanName}->connect("",$time);
                                }
                            }
                            else
                            {
                                print "\t\tNotice:\n\tInvalid record: $record\n";
                            }
                        }
                        elsif($record =~ ">\" changed name to \"")
                        {
                            #"Captain Salami<15>" changed name to "edward penishands<15>"
                            #$record =~ /^"(.+?)\<(\d+?)\>" changed name to "(.+?)\<(\d+?)\>"/;
                            
                            # Make sure we have all the elements that we need
                            #if (!$1 || !$2 || !$3 || !$4) { next; }
                            
                            #my($oldName,$pNum_old,$newName,$pNum_new) = ($1,$2,$3,$4);
                            my($oldName,$pNum_old,$wonId,$team) = getPlayerInfo($player);
                            my($pNum_new,$newName) = ($pNum_old,$noun1);

                            if(!$oldName || !$pNum_old || !$newName) { next; }

                            $names{cleanName($newName)} = $stats->{'live'}{$pNum_old};
                            
                            if($stats->{'live'}{$pNum_old} && length($stats->{'live'}{$pNum_old}) >= 1)
                            {
                                if($config{'store_method'} eq "name")
                                {
                                    $stats->{'live'}{$pNum_old} = cleanName($newName);
                                    
                                    if(!$stats->{'players'}{$stats->{'live'}{$pNum_old}}{'exists'})
                                    {
                                        $stats->{'players'}{$stats->{'live'}{$pNum_old}} = new Player();
                                        $stats->{'players'}{$stats->{'live'}{$pNum_old}}->setWonId($stats->{'live'}{$pNum_old});
                                        $stats->{'players'}{$stats->{'live'}{$pNum_old}}->recordStartTime($time);
                                        $stats->{'players'}{$stats->{'live'}{$pNum_old}}->setName($newName);
                                    }
                                }
                                else
                                {
                                    $stats->{'players'}{$stats->{'live'}{$pNum_old}}->setName($newName);
                                }
                                $stats->{'server'}{'general'}{'name_changes'} += 1;
                            }
                            else
                            {
                                print "\t\tNotice:\n\tInvalid record: $record\n";
                                print "\tNo wonId on name change?\n";
                            }
                        }
                        elsif($record =~ "Started map \"")
                        {
                            #Spawning server "gasworks"
                            #$record =~ /^Spawning server "(\w+?)"/;
                            $stats->{'live'}{'currentMap'} = $noun1;
                            $stats->{'server'}{'maps'}{$noun1}{'total'} += 1;
                            $stats->{'server'}{'general'}{'maps'} += 1;
                        }
                        elsif($record =~ ">\" committed suicide with ")
                        {
                            #"edward penishands<15>" killed self with rpg_rocket
                            #$record =~ /^"(.+?)\<(\d+?)\>" killed self with (\w+)/;
                            
                            # Make sure we have all the elements that we need
                            #if (!$1 || !$2 || !$3) { next; }
                            
                            my($name, $pNum, $wonid, $team, $weapon) = (getPlayerInfo($player), $noun1);
                            
                            if(!$name || !$pNum || !$weapon) { next; }

                            if($stats->{'live'}{'currentMap'} && 
                               $stats->{'live'}{$pNum} && 
                               length($stats->{'live'}{$pNum}) >= 1)
                            {           
                                $stats->{'players'}{$stats->{'live'}{$pNum}}->
                                    kill(\%{$stats->{'players'}{$stats->{'live'}{$pNum}}},$weapon,$stats->{'live'}{'currentMap'});
                            }
                            else
                            {
                                print "\t\tNotice:\n\tInvalid record: $record\n";
                            }
                            
                            if($stats->{'live'}{'currentMap'})
                            {
                                $stats->{'server'}{'maps'}{$stats->{'live'}{'currentMap'}}{'suicides'} += 1;
                            }
                            $stats->{'server'}{'general'}{'suicides'} += 1;
                        }
                        # We should never see this because world is treated as a normal person
                        #elsif($record =~ ">\" killed by world with ")
                        #{
                        #    #"edward penishands<15>" killed by world with worldspawn
                        #    $record =~ /^"(.+?)\<(\d+?)\>" killed by world with (\w+)/;
                        #    
                        #    my($name, $pNum, $weapon) = ($1,$2,$3);
                        #    
                        #    # Make sure we have all the elements that we need
                        #    if (!$1 || !$2 || !$3) { next; }
   
                        #   if($stats->{'live'}{'currentMap'} && 
                        #       $stats->{'live'}{$pNum} && 
                        #       length($stats->{'live'}{$pNum}) >= 1)
                        #    {
                        ##        $stats->{'players'}{$stats->{'live'}{$pNum}}->
                         #           kill(\%{$stats->{'players'}{$stats->{'live'}{$pNum}}},$weapon,$stats->{'live'}{'currentMap'});
                         #   }
                         #   else
                         #   {
                         #       print "\t\tNotice:\n\tInvalid record: $record\n";
                         #   }
                         #   
                         #   if($stats->{'live'}{'currentMap'})
                         #   {
                         #       $stats->{'server'}{'maps'}{$stats->{'live'}{'currentMap'}}{'suicides'} += 1;
                         #   }
                         #   $stats->{'server'}{'general'}{'suicides'} += 1;
                        #}
                        elsif($record =~ ">\" joined team ")
                        {
                            #"<15>" changed to team penis
                            $stats->{'server'}{'general'}{'team_changes'} += 1;
                        }
                        elsif($config{'game_type'} eq "cstrike")
                        {
                           my($testname,$testpNum,$testwonId,$testteam) = getPlayerInfo($player);
                           if(!$stats->{'live'}{$testpNum}) 
                           {
                              next;
                           }
                            if($record =~ " joined team \"TERRORIST\"")
                            {
                                #*crimrOw*-[shOck]- is joining the TERRORIST force
                                #$record =~ /^(.+?) is joining the TERRORIST force/;
                                #my $wonId = $names{cleanName($1)};

                                my($name,$pNum,$wonId,$team) = getPlayerInfo($player);
                                $wonId = $stats->{'live'}{$pNum};

                                if($wonId && length($wonId) >= 1)
                                {
                                    $stats->{'players'}{$wonId}->joinT();
                                }
                                else
                                {
                                    print "\t\tNotice: WONID not matched for $1 (T join ignored)\n";
                                }
                            }
                            elsif($record =~ " joined team \"CT\"")
                            {
                                #$record =~ /^(.+?) is joining the CT force/;
                                my($name,$pNum,$wonId,$team) = getPlayerInfo($player);                                
                                $wonId = $stats->{'live'}{$pNum};

                                if($wonId && length($wonId) >= 1)
                                {
                                    $stats->{'players'}{$wonId}->joinCT();
                                }
                                else
                                {
                                    print "\t\tNotice: WONID not matched for $1 (CT join ignored)\n";
                                }
                            }
                            elsif($record =~ " triggered \"Touched_A_Hostage\"")
                            {
                                #j00b34t3r^ touched a hostage
                                #$record =~ /^(.+?) touched a hostage/;

                                my($name,$pNum,$wonId,$team) = getPlayerInfo($player);
                                $wonId = $stats->{'live'}{$pNum};

                                if($wonId && length($wonId) >= 1)
                                {
                                    $stats->{'players'}{$wonId}->touchHostage();
                                }
                                else
                                {
                                    print "\t\tNotice: WONID not matched for $1 (Hostage touch ignored\n";
                                }
                                $stats->{'server'}{'cstrike'}{'hostage_touches'} += 1;
                            }
                            elsif($record =~ " triggered \"Killed_A_Hostage\"")
                            {
                                #FunKiE killed a hostage
                                my($name,$pNum,$wonId,$team) = getPlayerInfo($player);
                                $wonId = $stats->{'live'}{$pNum};

                                if($wonId && length($wonId) >= 1)
                                {
                                    $stats->{'players'}{$wonId}->killHostage();
                                }
                                else
                                {
                                    print "\t\tNotice: WONID not matched for $1 (Hostage kill ignored)\n";
                                }
                                $stats->{'server'}{'cstrike'}{'hostage_kills'} += 1;            
                            }
                            elsif($record =~ "Team \"TERRORIST\" triggered \"Terrorists_Win\"")
                            {
                                #***#Terrorists_Win***
                                $stats->{'server'}{'cstrike'}{'T_wins'} += 1;
                            }
                            elsif($record =~ "Team \"CT\" triggered \"CTs_Win\"")
                            {
                                $stats->{'server'}{'cstrike'}{'CT_wins'} += 1;
                            }
                            elsif($record =~ "World triggered \"Round_Draw\"")
                            {
                                $stats->{'server'}{'cstrike'}{'draws'} += 1;
                            }
                            elsif($record =~ "Team \"TERRORIST\" triggered \"Hostages_Not_Rescued\"")
                            {
                                $stats->{'server'}{'cstrike'}{'T_wins'} += 1;
                            }
                            elsif($record =~ " triggered \"Planted_The_Bomb\"")
                            {
                                #Jack Ryan planted the bomb
                                #$record =~ /^(.+?) planted the bomb/;
                                my($name,$pNum,$wonId,$team) = getPlayerInfo($player);

                                my($useid) = $stats->{'live'}{$pNum};

                                if($useid && $stats->{'players'}{$useid})
                                {
                                    $stats->{'players'}{$useid}->plantBomb();
                                    $stats->{'live'}{'last_bomber'} = $useid;
                                }
                                else
                                {
                                    print "\t\tNotice: WONID not matched for $1 (Bomb plant ignored)\n";
                                }
                                $stats->{'server'}{'cstrike'}{'bomb_plants'} += 1;
                            }
                            elsif($record =~ "team \"TERRORIST\" triggered \"Target_Bombed\"")
                            {
                                if($stats->{'live'}{'last_bomber'})
                                {
                                    $stats->{'players'}{$stats->{'live'}{'last_bomber'}}->detBomb();
                                }
                                $stats->{'server'}{'cstrike'}{'bomb_dets'} += 1;
                            }
                            elsif($record =~ "Rescued_A_Hostage")
                            {
                                #GRIM REAPER has rescued a hostage
                               my($name,$pNum,$wonId,$team) = getPlayerInfo($player);
                               $wonId = $stats->{'live'}{$pNum};
                                if($wonId && length($wonId) >= 1)
                                {
                                    $stats->{'players'}{$wonId}->hRescue();
                                }
                                else
                                {
                                    print "\t\tNotice: WONID not matched for $1 (Hostage rescue ignored)\n";
                                }
                                $stats->{'server'}{'cstrike'}{'hostage_rescue'} += 1;
                            }
                            elsif($record =~ "Defused_The_Bomb")
                            {
                                #$record =~ /^(.+?) has defused the bomb/;
                                #my $wonId = $names{cleanName($1)};
                               my($name,$pNum,$wonId,$team) = getPlayerInfo($player);
                               $wonId = $stats->{'live'}{$pNum};

                                if($wonId && length($wonId) >= 1)
                                {
                                    $stats->{'players'}{$wonId}->defBomb();
                                }
                                $stats->{'server'}{'cstrike'}{'bomb_defuse'} += 1;
                            }
                            elsif($record =~ "Assasinated_The_VIP")
                            {
                                my($name,$pNum,$wonId,$team) = getPlayerInfo($player);
                                $wonId = $stats->{'live'}{$pNum};
                                if($wonId && length($wonId) >= 1)
                                {
                                    $stats->{'players'}{$wonId}->killVIP();
                                }
                                else
                                {
                                    print "\t\tNotice: WONID not matched for $1 (VIP assasination ignored)\n";
                                }
                                $stats->{'server'}{'cstrike'}{'VIP_kills'} += 1;
                            }
                            elsif($record =~ "Escaped_As_VIP")
                            {
                                $stats->{'server'}{'cstrike'}{'VIP_escapes'} += 1;
                            }
                        }
                        
                        $count++;
                        if($count % $config{'sleep_interval'} == 0) { sleep $config{'sleep_time'}; }
                    }
                }
                
                # Now that we've parsed this log, add its name to the old_logs structure
                #     so that we can ignore it in the future
                $stats->{'old_logs'}{$logfile} = 1;
                
            }
            else
            {
                # Log is not yet closed - tell the user that we're ignoring it
                print "\tIgnoring $logfile (unclosed)\n";
            }
        }
    }  

    $stats->{'server'}{'general'}{'lines'} += $count;
    undef @logs;
}

#------------------------------------------------------------------------------
# Converts a Half-Life log date/time string into epoch seconds
sub getSecondsFromLogFormat
{
    my $datetime = shift;
    my $epochseconds = 0;
  
    if ($datetime =~ /^L (\d\d)\/(\d\d)\/(\d\d\d\d) - (\d\d):(\d\d):(\d\d): (.+)/)
    {
        my ($mon, $day, $year, $hour, $min, $sec, $record);
        $mon = $1; $day = $2; $year = $3; $hour = $4; $min = $5; $sec = $6; $record = $7;
        $epochseconds = timelocal($sec, $min, $hour, $day, $mon - 1, $year - 1900);
    }

    return $epochseconds;
}

#------------------------------------------------------------------------------
# Create each of the main statistics HTML pages
sub createStatPages
{
    createMainPage();
    
    # Players have a few calculated value -- do it now so sorting on them is possible
    createPlayerStatSnapshot();

    # People like to sort by stat type so make 6 pages
    createRankingPage("skill");
    createRankingPage("kills");
    createRankingPage("deaths");
    createRankingPage("kpd");
    createRankingPage("kph");

    createPlayerPages();

    createAwardsPage();
    
    # This will clean out the players excess crap?
    my $key;
    foreach $key (@players)
    {
        if($key)
        {
            $stats->{'players'}{$key}->cleanMyself();
        }
    }

    writeHelpFile();

    $stats->{'server'}{'general'}{'last_time'} = calcTime($sTime);
    $stats->{'server'}{'general'}{'total_time'} += $stats->{'server'}{'general'}{'last_time'};
    createServerPage();
}

#------------------------------------------------------------------------------
# How we start an HTML table
sub tStart
{
    my ($text);
    $text = "<table width='750' border='0' cellspacing='1' cellpadding='5'>\n";
    return($text);
}

#------------------------------------------------------------------------------
# How we end an HTML table
sub tEnd
{
    return("</table>\n");
}

#------------------------------------------------------------------------------
# This is the doc header for every generated HTML page
sub pagehead
{
    my($text,$label,$pathPrefix);

    $label = shift;
    $pathPrefix = shift;
    
    $text = "<html><head><title>$label - $config{'server_name'} running $SCRIPT_NAME version $VERSION</title>\n";

    $text .= "<link rel='STYLESHEET' type='text/css' href='$pathPrefix$config{'css_color'}.css'>\n";
    $text .= "<style>\nBODY\n{\nmargin-top: 5;\nmargin-left: 20;\nbackground-color: #000000;\ntext: #eeeeee;\n}\n</style>\n</head>\n";
    $text .= "<body text='#eeeeee'><center>\n";
    $text .= banner($pathPrefix) . "<br><br>\n";

    $text .= "<table width='750' border='0' cellspacing='0' cellpadding='5'>\n";
    $text .= "<tr class='headerbg' valign='middle'>\n";

    $text .= "<td align='left' width='400'>\n";
    $text .= "<font class='header'> $label </font>\n";
    $text .= "</td>\n";

    $text .= "<td align='right'>\n";
    $text .= "<font class='normal'>jump to:\n";
    $text .= "<a href='$pathPrefix" . "index.html'>main</a> | <a href='$pathPrefix" . "awards.html'>awards</a> | <a href='$pathPrefix" . "skill.html'>players</a> | <a href='$pathPrefix" . "server.html'>server</a> | <a href='$pathPrefix" . "help.html'>help?</a> \n"; 
    $text .= "</font>\n";
    $text .= "</td>\n";
    
    $text .= "</tr>\n";
    $text .= "</table>\n";

    return($text);
}

#------------------------------------------------------------------------------
# This page banner goes on the top of every generated HTML page
sub banner
{
    my ($banner,$pathPrefix);
    $pathPrefix = shift;
    $banner = "<table><tr valign='bottom'><td align='center'>\n";
    $banner .= "<a href='$HOME_PAGE' target='_top'><img src='$config{'rel_img_dir'}/banner.jpg' border='0'></a>\n";
    $banner .= "</td></tr></table>\n";
    return($banner);
}

#------------------------------------------------------------------------------
# This page footer goes on the bottom of every generated HTML page
sub pagefoot
{
    my ($text);
    $text = "<br><br><a href='$HOME_PAGE' target='_blank'>$SCRIPT_NAME</a>\n";
    $text .= "<font class='small'> ($VERSION, $DATE) was developed by $AUTHOR_NAME</font>\n";
    $text .= "<br><font class='small'>Last execution date ";

    $text .= getTimeString(time);
    $text .= " </font></center></body></html>\n";
    return($text);
}

#------------------------------------------------------------------------------
# Change current time (in epoch seconds) into a human-readable string
sub getTimeString
{
    my $text;
    
    my ($sec,$min,$hour,$mday,$mon,$year,$wday,$yday,$isdst)=localtime(shift);
    $year += 1900;
    $mon += 1;

    $text .= sprintf("%02.0f/", $mon);
    $text .= sprintf("%02.0f/", $mday);
    $text .= sprintf("%02.0f ", $year);
    $text .= sprintf("%02.0f:", $hour);
    $text .= sprintf("%02.0f:", $min);
    $text .= sprintf("%02.0f", $sec);
    return($text);
}

#------------------------------------------------------------------------------
# Create the main table of contents HTML page
sub createMainPage
{
    print "\tCreating main HTML\n";
    
    open(FILE, ">$config{'abs_html_dir'}/index.html") || die $!;
    print FILE pagehead("Main Menu",""), "<br>\n";
    print FILE tStart();

    print FILE "<tr valign='middle' class='cell1'>\n";
    print FILE "<td align='center'>\n";
    print FILE "<a href='skill.html'><img src='$config{'rel_img_dir'}/players.gif' border='0'></a>\n";
    print FILE "</td>\n";
    print FILE "<td align='center'>\n";
    print FILE "<a href='skill.html'><b>Player Ranking</b></a>\n";
    print FILE "</td>\n";
    print FILE "<td align='left' width='350'>\n";
    print FILE "<font class='small'><blockquote>This section has a ranked listing of all players on this server.  You can click on a player's stats link to see specific information about that player</blockquote></font>\n";
    print FILE "</td>\n";
    print FILE "</tr>\n";

    print FILE "<tr valign='middle' class='cell2'>\n";
    print FILE "<td align='center'>\n";
    print FILE "<a href='awards.html'><img src='$config{'rel_img_dir'}/awards.gif' border='0'></a>\n";
    print FILE "</td>\n";
    print FILE "<td align='center'>\n";
    print FILE "<a href='awards.html'><b>Awards Listing</b></a>\n";
    print FILE "</td>\n";
    print FILE "<td align='left' width='350'>\n";
    print FILE "<font class='small'><blockquote>This section displays all the awards that halfStats creates.  You can view highest weapon usage, and various other award listings</blockquote></font>\n";
    print FILE "</td>\n";
    print FILE "</tr>\n";

    print FILE "<tr valign='middle' class='cell3'>\n";
    print FILE "<td align='center'>\n";
    print FILE "<a href='server.html'><img src='$config{'rel_img_dir'}/server.gif' border='0'></a>\n";
    print FILE "</td>\n";
    print FILE "<td align='center'>\n";
    print FILE "<a href='server.html'><b>Server Statistics</b></a>\n";
    print FILE "</td>\n";
    print FILE "<td align='left' width='350'>\n";
    print FILE "<font class='small'><blockquote>The server statistics section lists server-wide map, weapon and halfStats configuration data.</blockquote></font>\n";
    print FILE "</td>\n";
    print FILE "</tr>\n";

    print FILE "<tr valign='middle' class='cell4'>\n";
    print FILE "<td align='center'>\n";
    print FILE "<a href='help.html'><img src='$config{'rel_img_dir'}/help.gif' border='0'></a>\n";
    print FILE "</td>\n";
    print FILE "<td align='center'>\n";
    print FILE "<a href='help.html'><b>Questions?</b></a>\n";
    print FILE "</td>\n";
    print FILE "<td align='left' width='350'>\n";
    print FILE "<font class='small'><blockquote>If you are unclear on how to use halfStats or have a question about how certain things are calculated or why some data might seem incorrect, read this section as it has answers to many common questions.</blockquote></font>\n";
    print FILE "</td>\n";
    print FILE "</tr>\n";

    print FILE tEnd();
    print FILE pagefoot();
    close(FILE);
}

#------------------------------------------------------------------------------
# Creates the server statistics HTML page
sub createServerPage
{
    my($key,$count,$map,$weapon,$vict,$kph,$rank,$item,$itemCat,$mod,$mcount,$col,$colcount,$victim,$class);
    $count = 0;

    print "\tCreating server HTML\n";

    open(FILE, ">$config{'abs_html_dir'}/server.html") || die $!;
    print FILE pagehead("Server Statistics","");

    print FILE "<br>", tStart();
    print FILE "<tr class='headerbg'><td><font class='header'>halfStats related statistics</font></td></tr>\n";
    print FILE tEnd();
       
    print FILE tStart();


    print FILE "<tr>\n";
    print FILE "<td class='cell1' width='364'><font class='normal'>Server admin</font></td>";
    print FILE "<td class='cell2' width='363'><font class='normal'><a href='mailto:" . $config{'admin_email'} . "'>";
    print FILE $config{'admin_name'} . "</a></font></td>\n";
    print FILE "</tr>\n";

    print FILE "<tr>\n";
    print FILE "<td class='cell1'><font class='normal'>Minimum kills to be ranked</font></td>";
    print FILE "<td class='cell2'><font class='normal'>" . $config{'minimum_kills'} . "</font></td>\n";
    print FILE "</tr>\n";

    print FILE "<tr>\n";
    print FILE "<td class='cell1'><font class='normal'>Minimum skill to be ranked</font></td>";
    print FILE "<td class='cell2'><font class='normal'>" . $config{'minimum_skill'} . "</font></td>\n";
    print FILE "</tr>\n";

    print FILE "<tr>\n";
    print FILE "<td class='cell1'><font class='normal'>Minimum visit frequency (in days) to be ranked</font></td>";
    print FILE "<td class='cell2'><font class='normal'>" . $config{'minimum_visit_frequency'} . "</font></td>\n";
    print FILE "</tr>\n";

    print FILE "<tr>\n";
    print FILE "<td class='cell1'><font class='normal'>Minimum play time (in minutes) to be ranked</font></td>";
    print FILE "<td class='cell2'><font class='normal'>" . $config{'minimum_minutes_played'} . "</font></td>\n";
    print FILE "</tr>\n";

    print FILE "<tr>\n";
    print FILE "<td class='cell1'><font class='normal'>Last execution date</font></td>";
    print FILE "<td class='cell2'><font class='normal'>" . getTimeString(time) . "</font></td>\n";
    print FILE "</tr>\n";

    print FILE "<tr>\n";
    print FILE "<td class='cell1'><font class='normal'>Execution run time</font></td>";
    print FILE "<td class='cell2'><font class='normal'>",$stats->{'server'}{'general'}{'last_time'}," seconds</font></td>\n";
    print FILE "</tr>\n";

    print FILE "<tr>\n";
    print FILE "<td class='cell1'><font class='normal'>Total lines read from logs</font></td>";
    print FILE "<td class='cell2'><font class='normal'>",$stats->{'server'}{'general'}{'lines'},"</font></td>\n";
    print FILE "</tr>\n";    

    print FILE "<tr>\n";
    print FILE "<td class='cell1'><font class='normal'>Size of logs parsed</font></td>";
    print FILE "<td class='cell2'><font class='normal'>",sprintf("%0.2f", (($stats->{'server'}{'general'}{'log_bytes'}/1024)/1024)),"MB</font></td>\n";
    print FILE "</tr>\n";    

    print FILE tEnd(), "<br>\n";

    print FILE tStart();
    print FILE "<tr class='headerbg'><td><font class='header'>Server wide statistics</font></td></tr>\n";
    print FILE tEnd();

    print FILE tStart();
    print FILE "<tr>\n";
    print FILE "<td class='cell1' width='364'><font class='normal'>Total connections</font></td>";
    print FILE "<td class='cell2' width='363'><font class='normal'>",$stats->{'server'}{'general'}{'connects'},"</font></td>\n";
    print FILE "</tr>\n";    

    print FILE "<tr>\n";
    print FILE "<td class='cell1'><font class='normal'>Total unique players</font></td>";
    print FILE "<td class='cell2'><font class='normal'>",$stats->{'server'}{'general'}{'unique'},"</font></td>\n";
    print FILE "</tr>\n";

    print FILE "<tr>\n";
    print FILE "<td class='cell1'><font class='normal'>Total map changes</font></td>";
    print FILE "<td class='cell2'><font class='normal'>",$stats->{'server'}{'general'}{'maps'},"</font></td>\n";
    print FILE "</tr>\n";

    print FILE "<tr>\n";
    print FILE "<td class='cell1'><font class='normal'>Total kills</font></td>";
    print FILE "<td class='cell2'><font class='normal'>",$stats->{'server'}{'general'}{'kills'},"</font></td>\n";
    print FILE "</tr>\n";

    print FILE "<tr>\n";
    print FILE "<td class='cell1'><font class='normal'>Total deaths</font></td>";
    print FILE "<td class='cell2'><font class='normal'>",$stats->{'server'}{'general'}{'kills'}+$stats->{'server'}{'general'}{'suicides'},"</font></td>\n";
    print FILE "</tr>\n";

    print FILE "<tr>\n";
    print FILE "<td class='cell1'><font class='normal'>Total suicides</font></td>";
    print FILE "<td class='cell2'><font class='normal'>",$stats->{'server'}{'general'}{'suicides'},"</font></td>\n";
    print FILE "</tr>\n";

    print FILE "<tr>\n";
    print FILE "<td class='cell1'><font class='normal'>Total name changes</font></td>";
    print FILE "<td class='cell2'><font class='normal'>",$stats->{'server'}{'general'}{'name_changes'},"</font></td>\n";
    print FILE "</tr>\n";

    if($config{'game_type'} eq "dm")
    {
        print FILE "<tr>\n";
        print FILE "<td class='cell1'><font class='normal'>Total team changes</font></td>";
        print FILE "<td class='cell2'><font class='normal'>",$stats->{'server'}{'general'}{'team_changes'},"</font></td>\n";
        print FILE "</tr>\n";
    }
    
    print FILE tEnd();

    if($config{'game_type'} eq "cstrike")
    {
        print FILE "<br>\n";
    
        print FILE tStart();
        print FILE "<tr class='headerbg'><td><font class='header'>server statistics for counter-strike</font></td></tr>\n";
        print FILE tEnd();
    
        print FILE tStart();
        
        if($stats->{'server'}{'cstrike'}{'CT_wins'})
        {
            print FILE "<tr>\n";
            print FILE "<td class='cell1' width='364'><font class='normal'>Total counter terrorist wins</font></td>";
            print FILE "<td class='cell2' width='363'><font class='normal'>",$stats->{'server'}{'cstrike'}{'CT_wins'},"</font></td>\n";
            print FILE "</tr>\n";
        }
    
        if($stats->{'server'}{'cstrike'}{'T_wins'})
        {
            print FILE "<tr>\n";
            print FILE "<td class='cell1'><font class='normal'>Total terrorist wins</font></td>";
            print FILE "<td class='cell2'><font class='normal'>",$stats->{'server'}{'cstrike'}{'T_wins'},"</font></td>\n";
            print FILE "</tr>\n";
        }
    
        if($stats->{'server'}{'cstrike'}{'draws'})
        {
            print FILE "<tr>\n";
            print FILE "<td class='cell1'><font class='normal'>Total draws</font></td>";
            print FILE "<td class='cell2'><font class='normal'>",$stats->{'server'}{'cstrike'}{'draws'},"</font></td>\n";
            print FILE "</tr>\n";
        }
    
        if($stats->{'server'}{'cstrike'}{'bomb_plants'})
        {
            print FILE "<tr>\n";
            print FILE "<td class='cell1'><font class='normal'>Total bombs planted</font></td>";
            print FILE "<td class='cell2'><font class='normal'>",$stats->{'server'}{'cstrike'}{'bomb_plants'},"</font></td>\n";
            print FILE "</tr>\n";
        }
    
        if($stats->{'server'}{'cstrike'}{'bomb_dets'})
        {
            print FILE "<tr>\n";
            print FILE "<td class='cell1'><font class='normal'>Total bombs detonated</font></td>";
            print FILE "<td class='cell2'><font class='normal'>",$stats->{'server'}{'cstrike'}{'bomb_dets'},"</font></td>\n";
            print FILE "</tr>\n";
        }
    
        if($stats->{'server'}{'cstrike'}{'bomb_defuse'})
        {
            print FILE "<tr>\n";
            print FILE "<td class='cell1'><font class='normal'>Total bombs defused</font></td>";
            print FILE "<td class='cell2'><font class='normal'>",$stats->{'server'}{'cstrike'}{'bomb_defuse'},"</font></td>\n";
            print FILE "</tr>\n";
        }
    
        if($stats->{'server'}{'cstrike'}{'hostage_touches'})
        {
            print FILE "<tr>\n";
            print FILE "<td class='cell1'><font class='normal'>Total hostages touched</font></td>";
            print FILE "<td class='cell2'><font class='normal'>",$stats->{'server'}{'cstrike'}{'hostage_touches'},"</font></td>\n";
            print FILE "</tr>\n";
        }
    
        if($stats->{'server'}{'cstrike'}{'hostage_rescue'})
        {
            print FILE "<tr>\n";
            print FILE "<td class='cell1'><font class='normal'>Total hostages rescued</font></td>";
            print FILE "<td class='cell2'><font class='normal'>",$stats->{'server'}{'cstrike'}{'hostage_rescue'},"</font></td>\n";
            print FILE "</tr>\n";
        }
    
        if($stats->{'server'}{'cstrike'}{'hostage_kills'})
        {
            print FILE "<tr>\n";
            print FILE "<td class='cell1'><font class='normal'>Total hostages killed</font></td>";
            print FILE "<td class='cell2'><font class='normal'>",$stats->{'server'}{'cstrike'}{'hostage_kills'},"</font></td>\n";
            print FILE "</tr>\n";
        }
    
        if($stats->{'server'}{'cstrike'}{'VIP_kills'})
        {
            print FILE "<tr>\n";
            print FILE "<td class='cell1'><font class='normal'>Total vip assasinations</font></td>";
            print FILE "<td class='cell2'><font class='normal'>",$stats->{'server'}{'cstrike'}{'VIP_kills'},"</font></td>\n";
            print FILE "</tr>\n";
        }
    
        if($stats->{'server'}{'cstrike'}{'VIP_escapes'})
        {
            print FILE "<tr>\n";
            print FILE "<td class='cell1'><font class='normal'>Total vip escapes</font></td>";
            print FILE "<td class='cell2'><font class='normal'>",$stats->{'server'}{'cstrike'}{'VIP_escapes'},"</font></td>\n";
            print FILE "</tr>\n";
        }
    
        print FILE tEnd();
    }
    
    print FILE "<br>\n";
    
    print FILE tStart();
    print FILE "<tr class='headerbg'><td><font class='header'>server statistics by map</font></td></tr>\n";
    print FILE tEnd();
    
    $rank = 1;
    
    print FILE tStart();
    foreach $map ( sort 
           { $stats->{'server'}{'maps'}{$b}{'total'} 
             <=> $stats->{'server'}{'maps'}{$a}{'total'} } 
           keys %{$stats->{'server'}{'maps'}} )
    {
        if($map)
        {
            print FILE "<tr>\n";
            print FILE "<td class='cell1' align='center' width='364'><font class='normal'><b>$map</b> [ ranked $rank ]\n";
            if(-e "$config{'abs_img_dir'}/maps/$config{'game_type'}/$map.jpg")
            {
                if ( $config{'hlmaps'} eq 'true' && $config{'hlmaps_url'} )
                {
                    print FILE "<br><A HREF=\"$config{'hlmaps_url'}?map=$map\"><IMG SRC=\"$config{'rel_img_dir'}/maps/$config{'game_type'}/$map.jpg\" ALT=\"$map\" WIDTH=212 HEIGHT=160 BORDER=0></A>";
                }
                else
                {
                    print FILE "<br><img src='$config{'rel_img_dir'}/maps/$config{'game_type'}/$map.jpg' ALT=\"$map\" WIDTH=212 HEIGHT=160 BORDER=0>\n";
                }
            }
            print FILE "</font></td>\n";
            print FILE "<td class='cell2'align='center' width='363'><table width='360' border='0' cellspacing='2' cellpading='5'>\n";
            
            if($rank ==  1)
            {
                print FILE "<font class='small'><b>Total uses (% of total)</b></font><br>";
            }
            
            print FILE "<tr valign='top'>\n";
            print FILE "<td class='cell3' width='75'><font class='normal'>Kills:</td>\n";
            print FILE "<td class='cell3' width='259'>", getBar("",$stats->{'server'}{'maps'}{$map}{'kills'},$stats->{'server'}{'general'}{'kills'}), "</font>\n";
            print FILE "</td></tr>\n";
            
            print FILE "<tr valign='top'>\n";
            print FILE "<td class='cell4'><font class='normal'>Suicides:</td>\n";
            print FILE "<td class='cell4'>", getBar("",$stats->{'server'}{'maps'}{$map}{'suicides'},$stats->{'server'}{'general'}{'suicides'}), "</font>\n";
            print FILE "</td></tr>\n";
            
            print FILE "<tr valign='top'>\n";
            print FILE "<td class='cell3'><font class='normal'>Times played:</td>\n";
            print FILE "<td class='cell3'>", getBar("",$stats->{'server'}{'maps'}{$map}{'total'},$stats->{'server'}{'general'}{'maps'}),"</font>\n";
            print FILE "</td></tr>\n";
            print FILE "</table></td>\n";
            print FILE "</tr>\n";
            $rank++;
            if($rank % $config{'sleep_interval'} == 0) { sleep $config{'sleep_time'}; }
        }
    }
    print FILE tEnd(), "<br>\n";
    
    print FILE tStart();
    print FILE "<tr class='headerbg'><td><font class='header'>server statistics by weapon</font></td></tr>\n";
    print FILE tEnd();
    
    $rank = 1;

    print FILE tStart();
    foreach $weapon ( sort
              { $stats->{'server'}{'weapons'}{$b}
            <=> $stats->{'server'}{'weapons'}{$a} }
              keys %{$stats->{'server'}{'weapons'}} )
    {
        print FILE "<tr height='75'>\n";
        print FILE "<td class='cell1' align='center' width='364'>\n";
        if(-e "$config{'abs_img_dir'}/weapons/$config{'game_type'}/$weapon.jpg")
        {
            print FILE "<img src='$config{'rel_img_dir'}/weapons/$config{'game_type'}/$weapon.jpg' border='0'>\n";
        }
        else
        {
            print FILE "<font class='normal'><b>$weapon</b></font>\n";
        }
        print FILE "</td>\n";
        
        print FILE "<td class='cell2' align='center' width='363'>\n";
        
        print FILE "<table width='360' class='cell1' border='0' cellspacing='2' cellpadding='5'><tr>\n";
        print FILE "<td class='small' align='left' width='186'>[ <font class='small'>ranked # $rank</font> ]</td>\n";
        print FILE "<td class='small' align='left' width='148'>[ <font class='small'>$weapon</font> ]</td>\n";
        print FILE "</tr></table>\n";
        
        print FILE "<table width='360' class='cell2' border='0' cellspacing='2' cellpadding='5'><tr>\n";
        
        print FILE "<td align='left' class='cell3' width='86'>\n";
        print FILE "<font class='normal'>$stats->{'server'}{'weapons'}{$weapon}</font>\n";
        print FILE "</td>\n";
        
        print FILE "<td align='left' class='cell3' width='88'>\n";
        print FILE "<font class='normal'>";
        print FILE sprintf("%0.1f",($stats->{'server'}{'weapons'}{$weapon}/$stats->{'server'}{'general'}{'kills'})*100);
        print FILE "% </font>\n";
        print FILE "</td>\n";       
        
        print FILE "<td align='left' class='cell3' width='148'>";
        print FILE getBar("",$stats->{'server'}{'weapons'}{$weapon},$stats->{'server'}{'general'}{'kills'},"no text");
        print FILE "</td>\n";
        
        print FILE "</tr></table>\n";
        
        print FILE "</td>\n";
        print FILE "</tr>\n";
        $rank++;
        if($rank % $config{'sleep_interval'} == 0) { sleep $config{'sleep_time'}; }
    }
    print FILE tEnd(), "<br>\n";
    
    print FILE pagefoot();
    close(FILE);
    
}

#------------------------------------------------------------------------------
# Calculates every player's kpd and kph statistics
sub createPlayerStatSnapshot
{
    my($key);
    foreach $key(@players)
    {
        # Calculate kills per death
        if($stats->{'players'}{$key}->{'deaths'} > 0) 
        { 
            $stats->{'players'}{$key}->{'kpd'} 
            = sprintf("%0.2f", $stats->{'players'}{$key}->{'kills'} / $stats->{'players'}{$key}->{'deaths'}); 
        }
        else
        {
            $stats->{'players'}{$key}->{'kpd'} = $stats->{'players'}{$key}->{'kills'};
        }
    
        # Calculate kills per hour
        if($stats->{'players'}{$key}->{'total_time'} > 60)
        {
            $stats->{'players'}{$key}->{'kph'}
                = sprintf("%0.0f", 
                  ($stats->{'players'}{$key}->{'kills'} / (($stats->{'players'}{$key}->{'total_time'}/60)/60)));
        }
        else
        {
            $stats->{'players'}{$key}->{'kph'} = "0";
        }

    }
}

#------------------------------------------------------------------------------
# Create the main player ranking HTML page
sub createRankingPage
{    
    my $sortby = shift;

    print "\tCreating player $sortby ranking HTML\n";

    open(FILE, ">$config{'abs_html_dir'}/$sortby.html") || die $!;
    print FILE pagehead("Player Ranking","");
    
    my ($key,$count,$kpd,$last,$kph,$class,$printName,$foundAward);

    $count = 1;$foundAward = 0;

    print FILE "<br>\n", tStart();
    
    print FILE "<tr>\n";
    print FILE "<td align='center' class='headerbg'><font class='normal'><b>Rank</b></font></td>\n";
    print FILE "<td align='center' class='headerbg'><font class='normal'><b>Name</b></font></td>\n";
    print FILE "<td align='center' class='headerbg'><font class='normal'><b>[ stat link ]</b></font></td>\n";

    if($sortby eq "kills") { $class = 'sortbg'; } else { $class = 'headerbg'; }
    print FILE "<td align='center' class='$class'><font class='normal'><b>\n";
    print FILE "<a href='kills.html'>Kills</a>\n";
    print FILE "</b></font></td>\n";

    if($sortby eq "deaths") { $class = 'sortbg'; } else { $class = 'headerbg'; }
    print FILE "<td align='center' class='$class'><font class='normal'><b>\n";
    print FILE "<a href='deaths.html'>Deaths</a>\n";
    print FILE "</b></font></td>\n";

    if($sortby eq "kpd") { $class = 'sortbg'; } else { $class = 'headerbg'; }
    print FILE "<td align='center' class='$class'><font class='normal'><b>\n";
    print FILE "<a href='kpd.html'>Efficiency</a>\n";
    print FILE "</b></font></td>\n";

    if($sortby eq "kph") { $class = 'sortbg'; } else { $class = 'headerbg'; }
    print FILE "<td align='center' class='$class'><font class='normal'><b>\n";
    print FILE "<a href='kph.html'>Kills/Hr</a>\n";
    print FILE "</b></font></td>\n";

    if($sortby eq "skill") { $class = 'sortbg'; } else { $class = 'headerbg'; }
    print FILE "<td align='center' class='$class'><font class='normal'><b>\n";
    print FILE "<a href='skill.html'>Skill</a>\n";
    print FILE "</b></font></td>\n";

    print FILE "</tr>\n";

    $last = "cell2";

    foreach $key(sort 
         { $stats->{'players'}{$b}->{$sortby} <=> $stats->{'players'}{$a}->{$sortby} } 
         @players)
    {
        if($count == 1)
        {
            $awards{$sortby}{'wonId'} = $key;
            $awards{$sortby}{'value'} = $stats->{'players'}{$key}->{$sortby};
        }
    
        if($count < $config{'rank_max'} && canReportPlayer($key))
        {       
            
            if($last eq "cell1") { $last = "cell2"; } else { $last = "cell1"; }
            if(length($stats->{'players'}{$key}->{'name_html'}) > 30)
            {
                $stats->{'players'}{$key}->{'name_html'} 
                = substr($stats->{'players'}{$key}->{'name_html'}, 0, 30) . "...";
            }
    
            print FILE "<tr class='$last'>\n";
            print FILE "<td><font class='normal'><b>$count</b></font></td>\n";
            print FILE "<td><font class='normal'>", $stats->{'players'}{$key}->{'name_html'} ,"</font></td>\n";
            print FILE "<td align='center'><font class='normal'>[ <a href='players/", $key, ".html'>stats</a> ]</font></td>\n";
            print FILE "<td align='center'><font class='normal'>", $stats->{'players'}{$key}->{'kills'}, "</font></td>\n";
            print FILE "<td align='center'><font class='normal'>", $stats->{'players'}{$key}->{'deaths'}, "</font></td>\n";
            print FILE "<td align='center'><font class='normal'>", $stats->{'players'}{$key}->{'kpd'}, "</font></td>\n";
            print FILE "<td align='center'><font class='normal'>", $stats->{'players'}{$key}->{'kph'}, "</font></td>\n";
            print FILE "<td align='center'><font class='normal'>", $stats->{'players'}{$key}->{'skill'}, "</font></td>\n";
            print FILE "</tr>\n";
    
            $count += 1;
            if($count % $config{'sleep_interval'} == 0) { sleep $config{'sleep_time'}; }
        }
    }

    print FILE tEnd();

    print FILE pagefoot();
    close(FILE);

}

#------------------------------------------------------------------------------
# Create each player's HTML statistics page
sub createPlayerPages
{
    my($key,$count,$map,$weapon,$vict,$kph,$rank,$mod,$mcount,$col,$colcount,$victim,$class,$alias);
    
    $count = 0;

    print "\tCreating player pages HTML\n";
    foreach $key (@players)
    {
        if (length($key) < 1)       { next; }
        if (!canReportPlayer($key)) { next; }
   
        #print $key, "\n\t-->", length($key), "\n";
        open(FILE, ">$config{'abs_html_dir'}/players/" . $key . ".html") || die $!;
        print FILE pagehead("individual statistics for " . $stats->{'players'}{$key}->{'name_html'},"../");
    
        print FILE tStart();
    
        print FILE "<tr>\n";
        print FILE "<td class='cell1' width='364'><font class='normal'>Kills</font></td>";
        print FILE "<td class='cell2' width='363'><font class='normal'>",$stats->{'players'}{$key}->{'kills'},"</font></td>\n";
        print FILE "</tr>\n";
    
        print FILE "<tr>\n";
        print FILE "<td class='cell1'><font class='normal'>Deaths</font></td>";
        print FILE "<td class='cell2'><font class='normal'>",$stats->{'players'}{$key}->{'deaths'},"</font></td>\n";
        print FILE "</tr>\n";
    
        print FILE "<tr>\n";
        print FILE "<td class='cell1'><font class='normal'>Suicides</font></td>";
        print FILE "<td class='cell2'><font class='normal'>",$stats->{'players'}{$key}->{'suicides'},"</font></td>\n";
        print FILE "</tr>\n";
    
        if($stats->{'players'}{$key}->{'deaths'} > 0)
        {
            print FILE "<tr>\n";
            print FILE "<td class='cell1'><font class='normal'>Kills/Deaths [efficiency]</font></td>";
            print FILE "<td class='cell2'><font class='normal'>",sprintf("%0.2f", $stats->{'players'}{$key}->{'kills'}/$stats->{'players'}{$key}->{'deaths'}),"</font></td>\n";
            print FILE "</tr>\n";
        }
    
        print FILE "<tr>\n";
        print FILE "<td class='cell1'><font class='normal'>Skill Rating</font></td>";
        print FILE "<td class='cell2'><font class='normal'>",$stats->{'players'}{$key}->{'skill'},"</font></td>\n";
        print FILE "</tr>\n";
    
        print FILE "<tr>\n";
        print FILE "<td class='cell1'><font class='normal'>Total Connections</font></td>";
        print FILE "<td class='cell2'><font class='normal'>",$stats->{'players'}{$key}->{'disconnects'},"</font></td>\n";
        print FILE "</tr>\n";
    
        if($stats->{'players'}{$key}->{'total_time'} > 0)
        {
          $kph = sprintf("%.2f", 
                   ( $stats->{'players'}{$key}->{'kills'} / 
                    ( ($stats->{'players'}{$key}->{'total_time'} / 3600) ) ));
            print FILE "<tr>\n";
            print FILE "<td class='cell1'><font class='normal'>Kills per Hour</font></td>";
            print FILE "<td class='cell2'><font class='normal'>$kph</font></td>\n";
            print FILE "</tr>\n";
        }
    
        print FILE "<tr>\n";
        print FILE "<td class='cell1'><font class='normal'>Total Time Played</font></td>";
        print FILE "<td class='cell2'><font class='normal'>";
        if($stats->{'players'}{$key}->{'total_time'} > 60)
        {
            print FILE sprintf("%0.0f", $stats->{'players'}{$key}->{'total_time'}/60)," minutes";
        }
        else
        {
            print FILE sprintf("%0.0f", $stats->{'players'}{$key}->{'total_time'})," seconds";
        }
    
        print FILE "</font></td>\n";
        print FILE "</tr>\n";
    
        if($config{'show_ip'} eq "true")
        {
            print FILE "<tr>\n";
            print FILE "<td class='cell1'><font class='normal'>Last IP [wonid]</font></td>";
            print FILE "<td class='cell2'><font class='normal'>",$stats->{'players'}{$key}->getIP()," [",$stats->{'players'}{$key}{'saved_wonId'},"] </font></td>\n";
            print FILE "</tr>\n";
    
        }
    
        print FILE "<tr>\n";
        print FILE "<td class='cell1'><font class='normal'>Last connection</font></td>";
        print FILE "<td class='cell2'><font class='normal'>",getTimeString($stats->{'players'}{$key}->getLastConnect()),"</font></td>\n";
        print FILE "</tr>\n";
    
        print FILE tEnd();
    
        print FILE "<br>\n";
            
        if($config{'game_type'} eq "cstrike")
        {
            print FILE tStart();
            print FILE "<tr class='headerbg'><td><font class='header'>Counter Strike Stats</font></td></tr>\n";
            print FILE tEnd();
    
            print FILE tStart();
    
                #'b_dets' => 0,'b_defuse' => 0,'hRescue' => 0,'b_plants' => 0,'CTjoins' => 0,'hostage_touch' => 0,'hKills' => 0
    
            if($stats->{'players'}{$key}->getBombPlants())
            {
                print FILE "<tr>\n";
                print FILE "<td class='cell1' width='364'><font class='normal'>Bombs planted</font></td>";
                print FILE "<td class='cell2' width='363'><font class='normal'>" . $stats->{'players'}{$key}->getBombPlants() . "</font></td>\n";
                print FILE "</tr>\n";
            }
            
            if($stats->{'players'}{$key}->getBombDets())
            {
                print FILE "<tr>\n";
                print FILE "<td class='cell1'><font class='normal'>Bombs detonated</font></td>";
                print FILE "<td class='cell2'><font class='normal'>" . $stats->{'players'}{$key}->getBombDets() . "</font></td>\n";
                print FILE "</tr>\n";
            }
    
            if($stats->{'players'}{$key}->getBombDefuse())
            {
                print FILE "<tr>\n";
                print FILE "<td class='cell1'><font class='normal'>Bombs defused</font></td>";
                print FILE "<td class='cell2'><font class='normal'>" . $stats->{'players'}{$key}->getBombDefuse() . "</font></td>\n";
                print FILE "</tr>\n";
            }
    
            if($stats->{'players'}{$key}->getHostageTouches())
            {
                print FILE "<tr>\n";
                print FILE "<td class='cell1'><font class='normal'>Hostages touched</font></td>";
                print FILE "<td class='cell2'><font class='normal'>" . $stats->{'players'}{$key}->getHostageTouches() . "</font></td>\n";
                print FILE "</tr>\n";
            }
    
            if($stats->{'players'}{$key}->getHRescue())
            {
                print FILE "<tr>\n";
                print FILE "<td class='cell1'><font class='normal'>Hostages rescued</font></td>";
                print FILE "<td class='cell2'><font class='normal'>" . $stats->{'players'}{$key}->getHRescue() . "</font></td>\n";
                print FILE "</tr>\n";
            }
    
            if($stats->{'players'}{$key}->getHostageKills())
            {
                print FILE "<tr>\n";
                print FILE "<td class='cell1'><font class='normal'>Hostages killed</font></td>";
                print FILE "<td class='cell2'><font class='normal'>" . $stats->{'players'}{$key}->getHostageKills() . "</font></td>\n";
                print FILE "</tr>\n";
            }
    
            if($stats->{'players'}{$key}->getVIPKills())
            {
                print FILE "<tr>\n";
                print FILE "<td class='cell1'><font class='normal'>VIP Assasinations</font></td>";
                print FILE "<td class='cell2'><font class='normal'>" . $stats->{'players'}{$key}->getVIPKills() . "</font></td>\n";
                print FILE "</tr>\n";
            }
    
            if($stats->{'players'}{$key}->getCTJoins())
            {
                print FILE "<tr>\n";
                print FILE "<td class='cell1'><font class='normal'>Times as a counter-terrorist</font></td>";
                print FILE "<td class='cell2'><font class='normal'>" . $stats->{'players'}{$key}->getCTJoins() . "</font></td>\n";
                print FILE "</tr>\n";
            }
    
            if($stats->{'players'}{$key}->getTJoins())
            {
                print FILE "<tr>\n";
                print FILE "<td class='cell1'><font class='normal'>Times as a terrorist</font></td>";
                print FILE "<td class='cell2'><font class='normal'>" . $stats->{'players'}{$key}->getTJoins() . "</font></td>\n";
                print FILE "</tr>\n";
            }
    
            print FILE tEnd();
            
            print FILE "<br>\n";
        }
    
        print FILE tStart();
        print FILE "<tr class='headerbg'><td><font class='header'>kill, death and suicide statistics by map</font></td></tr>\n";
        print FILE tEnd();
    
        $rank = 1;
    
        # Print out per-map stats
        print FILE tStart();
        foreach $map ( sort 
                   { $stats->{'players'}{$key}->{'maps'}{'kills'}{$b} 
                 <=> $stats->{'players'}{$key}->{'maps'}{'kills'}{$a} } 
                   keys %{$stats->{'players'}{$key}->{'maps'}{'kills'}} )
        {
            print FILE "<tr>\n";
            print FILE "<td class='cell1' align='center' width='364'><font class='normal'><b>$map</b> [ $rank ]\n";
            if(-e "$config{'abs_img_dir'}/maps/$config{'game_type'}/$map.jpg")
            {
                if ( $config{'hlmaps'} eq 'true' && $config{'hlmaps_url'} )
                {
                    print FILE "<br><A HREF=\"$config{'hlmaps_url'}?map=$map\"><img src='$config{'rel_img_dir'}/maps/$config{'game_type'}/$map.jpg' ALT=\"$map\" WIDTH=212 HEIGHT=160 BORDER=0></A>";
                }
                else
                {
                    print FILE "<br><img src='$config{'rel_img_dir'}/maps/$config{'game_type'}/$map.jpg' ALT=\"$map\" WIDTH=212 HEIGHT=160 BORDER=0>\n";
                }
            }
            print FILE "</font></td>\n";
            print FILE "<td class='cell2' align='center' width='363'><table width='360' border='0' cellspacing='2' cellpadding='5'>\n";
    
            if($rank ==  1)
            {
                print FILE "<font class='small'><b>Total uses (% of total)</b></font><br>";
            }
    
            print FILE "<tr valign='top'>\n";
            print FILE "<td class='cell3' width='78'><font class='normal'>Kills:</td>\n";
            print FILE "<td class='cell3' width='256'>", getBar("../",$stats->{'players'}{$key}->{'maps'}{'kills'}{$map},$stats->{'players'}{$key}->{'kills'}), "</font>\n";
            print FILE "</td></tr>\n";
    
            print FILE "<tr valign='top'>\n";
            print FILE "<td class='cell4'><font class='normal'>Deaths:</td>\n";
            print FILE "<td class='cell4'>", getBar("../",$stats->{'players'}{$key}->{'maps'}{'deaths'}{$map},$stats->{'players'}{$key}->{'deaths'}), "</font>\n";
            print FILE "</td></tr>\n";
    
            print FILE "<tr valign='top'>\n";
            print FILE "<td class='cell3'><font class='normal'>Suicides:</td>\n";
            print FILE "<td class='cell3'>", getBar("../",$stats->{'players'}{$key}->{'maps'}{'suicides'}{$map},$stats->{'players'}{$key}->{'suicides'}),"</font>\n";
            print FILE "</td></tr>\n";
            print FILE "</table></td>\n";
            print FILE "</tr>\n";
            $rank++;
            if($rank % $config{'sleep_interval'} == 0) { sleep $config{'sleep_time'}; }
        }
        print FILE tEnd(), "<br>\n";
    
        $rank = 1;
    
            print FILE tStart();
            print FILE "<tr class='headerbg'><td><font class='header'>kill statistics by weapon</font></td></tr>\n";
            print FILE tEnd();
    
            print FILE tStart();
            foreach $weapon ( sort
                           { $stats->{'players'}{$key}->{'weapons'}{$b}
                             <=> $stats->{'players'}{$key}->{'weapons'}{$a} }
                           keys %{$stats->{'players'}{$key}->{'weapons'}} )
            {
            print FILE "<tr height='75'>\n";
                print FILE "<td class='cell1' align='center' width='364'>\n";
                if(-e "$config{'abs_img_dir'}/weapons/$config{'game_type'}/$weapon.jpg")
                {
                    print FILE "<img src='$config{'rel_img_dir'}/weapons/$config{'game_type'}/$weapon.jpg' border='0'>\n";
                }
            else
            {
                print FILE "<font class='normal'><b>$weapon</b></font>\n";
            }
            print FILE "</td>\n";
    
            print FILE "<td class='cell2' align='center' width='363'>\n";
    
            print FILE "<table width='360' class='cell1' border='0' cellspacing='2' cellpadding='5'><tr>\n";
            print FILE "<td class='small' align='left' width='188'>[ <font class='small'>ranked # $rank</font> ]</td>\n";
            print FILE "<td class='small' align='left' width='146'>[ <font class='small'>$weapon</font> ]</td>\n";
            print FILE "</tr></table>\n";
            
            print FILE "<table width='360' class='cell2' border='0' cellspacing='2' cellpadding='5'><tr>\n";
    
            print FILE "<td align='left' class='cell3' width='88'>\n";
            print FILE "<font class='normal'>$stats->{'players'}{$key}->{'weapons'}{$weapon}</font>\n";
            print FILE "</td>\n";
    
            print FILE "<td align='left' class='cell3' width='88'>\n";
    
            print FILE "<font class='normal'>";
            print FILE sprintf("%0.1f",($stats->{'players'}{$key}->getWeaponUses($weapon)/$stats->{'players'}{$key}->getKills())*100);
            print FILE "% </font>\n";
            print FILE "</td>\n";       
    
            print FILE "<td align='left' class='cell3' width='146'>";
            print FILE getBar("../",$stats->{'players'}{$key}->{'weapons'}{$weapon},$stats->{'players'}{$key}->{'kills'},"no text");
            print FILE "</td>\n";
    
            print FILE "</tr></table>\n";
    
            print FILE "</td>\n";
            print FILE "</tr>\n";
            $rank++;
            if($rank % $config{'sleep_interval'} == 0) { sleep $config{'sleep_time'}; }
        }
        print FILE tEnd(), "<br>\n";
    
        print FILE tStart();
        print FILE "<tr class='headerbg'><td><font class='header'>statistics by opponent</font></td></tr>\n";
        print FILE tEnd();
    
        $rank = 1;
    
        # Print out per-victim stats
        print FILE tStart();
        print FILE "<tr class='headerbg'>\n";
        print FILE "<td class='small'><b>Rank</b></td>\n";
        print FILE "<td class='small'><b>Opponent</b></td>\n";
        print FILE "<td class='small'><b>[ stat link ]</b></td>\n";
        print FILE "<td class='small'><b>Kills vs</b></td>\n";
        print FILE "<td class='small'><b>Deaths vs</b></td>\n";
        print FILE "<td class='small'><b>Efficiency</b></td>\n";
        print FILE "<td class='small'><b>Favorite Weapon</b></td>\n";
        print FILE "</tr>\n";
        
        $class = "cell1";
    
        foreach $victim ( sort { $stats->{'players'}{$key}->getVictimKills($b) <=> $stats->{'players'}{$key}->getVictimKills($a) } keys %{$stats->{'players'}{$key}->{'victims'}} )
        {
            my(@favWeap,$vkpd);
            
            # We only list the opponent's stats if:
            #     1) they have an HTML file
            #     2) the killer has killed them at least once
            #     3) if they are displayable determined by user preferences
            if($victim && 
               $stats->{'players'}{$key}->getVictimKills($victim) > 0 &&
               length($stats->{'players'}{$victim}->{'name_html'}) >= 1 &&
               canReportPlayer($stats->{'players'}{$victim}->{'wonId'}) )
            {
                @favWeap = getFavWeapon($key,$victim);
                if($stats->{'players'}{$victim}->getVictimKills($key) > 0)
                {
                    $vkpd = sprintf("%0.2f", ($stats->{'players'}{$key}->{'victims'}{$victim}{'total'} /
                                  $stats->{'players'}{$victim}->{'victims'}{$key}{'total'}));
                }
                else
                {
                    $vkpd = $stats->{'players'}{$key}->{'victims'}{$victim}{'total'};
                }
        
                print FILE "<tr class='$class' valign='top'>\n";
                print FILE "<td class='normal' align='center'><b>$rank</b></td>";
                
                print FILE "<td class='small' align='center'>";
                print FILE $stats->{'players'}{$victim}->{'name_html'}; 
                print FILE "</td>\n";
                
                print FILE "<td class='small' align='center'>";

                print FILE "[ <a href='players/";
                print FILE $stats->{'players'}{$victim}->{'wonId'}, ".html'>stats</a> ]";
                print FILE "</td>\n";
                
                print FILE "<td class='normal' align='center'>";
                if($stats->{'players'}{$key}->{'victims'}{$victim}{'total'} > 0)
                {
                    print FILE $stats->{'players'}{$key}->{'victims'}{$victim}{'total'};
                }
                else
                {
                    print FILE "0";
                }
                print FILE "</td>\n";
        
                print FILE "<td class='normal' align='center'>";
                
                if($victim && $key)
                {
                    if($stats->{'players'}{$victim}{'victims'}{$key}{'total'})
                    {
                        print FILE $stats->{'players'}{$victim}->getVictimKills($key);
                    }
                    else
                    {
                        print FILE "0";
                    }
                   
                }
                else
                {
                    print FILE "0";
                }
                print FILE "</td>\n";
                
                print FILE "<td class='normal' align='center'>$vkpd</td>\n";
                
                print FILE "<td class='cell1' align='center'><table width='110'><tr valign='middle'><td align='left'>";
                if(-e $config{'abs_img_dir'} . "/weapons/$config{'game_type'}/" . $favWeap[0] . ".jpg")
                {
                    print FILE "<img src='$config{'rel_img_dir'}/weapons/$config{'game_type'}/" . 
                    $favWeap[0] . ".jpg' border='0' height='35' width='125'>\n";
                }
                else
                {
                    print FILE "<font class='small'>" . $favWeap[0] . "</font>\n";
                }
                print FILE "</td><td align='right'>\n";
                print FILE "<font class='contrast'><b>" . $favWeap[1] . "</b></font>\n";
                print FILE "</td></tr></table></td>\n"; 
                
                print FILE "</tr>\n";
                $rank++;
                if($rank % $config{'sleep_interval'} == 0) { sleep $config{'sleep_time'}; }
            }
        }
    
        print FILE tEnd(), "<br>\n";

        print FILE tStart();
        print FILE "<tr class='headerbg'><td><font class='header'>this players aliases</font></td></tr>\n";
        print FILE tEnd();
    
        $rank = 1;
    
        print FILE tStart();
    
        print FILE "<tr class='headerbg'>\n";
        print FILE "<td class='small' width='49'><b>Rank</b></td>\n";
        print FILE "<td class='small' width='188'><b>Name</b></td>\n";
        print FILE "<td class='small' width='89'><b>Times Used</b></td>\n";
        print FILE "</tr>\n";
    
        foreach $alias (sort 
                { 
                    $stats->{'players'}{$key}->{'previous_names'}{$b} <=> 
                    $stats->{'players'}{$key}->{'previous_names'}{$a} 
                } keys %{$stats->{'players'}{$key}->{'previous_names'}} )
        {        
            print FILE "<tr class='normal'>\n";
            print FILE "<td class='cell1'>$rank</td>\n";
            print FILE "<td class='cell2'>$alias</td>\n";
            print FILE "<td class='cell3'>" . $stats->{'players'}{$key}->{'previous_names'}{$alias} . "</td>\n";
            print FILE "</tr>\n";
            $rank += 1;
        }
    
        print FILE tEnd(), "<br>\n";
    
        print FILE pagefoot();
        close(FILE);
    
        $count++;
        if($count % $config{'sleep_interval'} == 0){sleep $config{'sleep_time'};}
    }
}

#------------------------------------------------------------------------------
# Create the main awards HTML page
sub createAwardsPage
{
    my($key,$count,$map,$weapon,$rank,$mod,$mcount,$col,$colcount,$victim,$class,$high);

    print "\tCreating awards HTML\n";

    open(FILE, ">$config{'abs_html_dir'}/awards.html") || die $!;
    print FILE pagehead("Award Listing","");

    $rank = 1;

    print FILE "<br>", tStart();
    print FILE "<tr class='headerbg'><td><font class='header'>Awards for players that excel</font></td></tr>\n";
    print FILE tEnd();

    print FILE tStart();

    print FILE "<tr class='normal'>\n";
    print FILE "<td class='cell1' align='left' width='363'><font class='white'><b>Highest skill</b></font></td>\n";
    print FILE "<td class='cell2' align='left' width='114'>" . $awards{'skill'}{'value'} . "</td>\n";
    print FILE "<td class='cell3' align='right' width='114'><font class='contrast'>" . 
            $stats->{'players'}{$awards{'skill'}{'wonId'}}->{'name_html'} . "</font></td>\n";
    print FILE "<td class='cell3' align='center' width='114'>[ ";
    print FILE "<a href='players/$awards{'skill'}{'wonId'}.html'>stats</a> ]</td>\n</tr>\n";
    
    print FILE "<tr class='normal'>\n";
    print FILE "<td class='cell1' align='left'><font class='white'><b>Fastest killer</b></font></td>\n";
    print FILE "<td class='cell2' align='left'>" . $awards{'kph'}{'value'} . " kills/hr</td>\n";
    print FILE "<td class='cell3' align='right'><font class='contrast'>" . 
            $stats->{'players'}{$awards{'kph'}{'wonId'}}->{'name_html'} . "</font></td>\n";
    print FILE "<td class='cell3' align='center'>[ ";
    print FILE "<a href='players/$awards{'kph'}{'wonId'}.html'>stats</a> ]</td>\n</tr>\n";

    print FILE "<tr class='normal'>\n";
    print FILE "<td class='cell1' align='left'><font class='white'><b>Most efficient killer</b></font></td>\n";
    print FILE "<td class='cell2' align='left'>" . $awards{'kpd'}{'value'} . " kills/death</td>\n";
    print FILE "<td class='cell3' align='right'><font class='contrast'>" . 
            $stats->{'players'}{$awards{'kpd'}{'wonId'}}->{'name_html'} . "</font></td>\n";
    print FILE "<td class='cell3' align='center'>[ ";
    print FILE "<a href='players/$awards{'kpd'}{'wonId'}.html'>stats</a> ]</td>\n</tr>\n";

    print FILE "<tr class='normal'>\n";
    print FILE "<td class='cell1' align='left'><font class='white'><b>Highest volume killer</b></font></td>\n";
    print FILE "<td class='cell2' align='left'>" . $awards{'kills'}{'value'} . " kills total</td>\n";
    print FILE "<td class='cell3' align='right'><font class='contrast'>" . 
            $stats->{'players'}{$awards{'kills'}{'wonId'}}->{'name_html'} . "</font></td>\n";
    print FILE "<td class='cell3' align='center'>[ ";
    print FILE "<a href='players/$awards{'kills'}{'wonId'}.html'>stats</a> ]</td>\n</tr>\n";

    print FILE "<tr class='normal'>\n";
    print FILE "<td class='cell1' align='left'><font class='white'><b>Most likely to get shot</b></font></td>\n";
    print FILE "<td class='cell2' align='left'>" . $awards{'deaths'}{'value'} . " deaths total</td>\n";
    print FILE "<td class='cell3' align='right'><font class='contrast'>" .
            $stats->{'players'}{$awards{'deaths'}{'wonId'}}->{'name_html'} . "</font></td>\n";
    print FILE "<td class='cell3' align='center'>[ ";
    print FILE "<a href='players/$awards{'deaths'}{'wonId'}.html'>stats</a> ]</td>\n</tr>\n";

    $high = getSuicideAward();

    if($high->{'wonId'})
    {
       print FILE "<tr class='normal'>\n";
       print FILE "<td class='cell1' align='left'><font class='white'><b>Most likely to off themself</b></font></td>\n";
       print FILE "<td class='cell2' align='left'>" . $high->{'high'} . " suicides</td>\n";
       print FILE "<td class='cell3' align='right'><font class='contrast'>" .
               $stats->{'players'}{$high->{'wonId'}}->{'name_html'} . "</font></td>\n";
       print FILE "<td class='cell3' align='center'>[ ";

       print FILE "<a href='players/$high->{'wonId'}.html'>stats</a> ]</td>\n</tr>\n";
    }

    $high = getTimeAward();

    if($high->{'wonId'})
    {
       print FILE "<tr class='normal'>\n";
       print FILE "<td class='cell1' align='left'><font class='white'><b>Most time on the server</b></font></td>\n";
       print FILE "<td class='cell2' align='left'>" . sprintf("%0.2f", $high->{'high'}/60) . " minutes</td>\n";
       print FILE "<td class='cell3' align='right'><font class='contrast'>" .
               $stats->{'players'}{$high->{'wonId'}}->{'name_html'} . "</font></td>\n";
       print FILE "<td class='cell3' align='center'>[ ";
       print FILE "<a href='players/$high->{'wonId'}.html'>stats</a> ]</td>\n</tr>\n";
    }

    $high = getConnectAward();

    if($high->{'wonId'})
    {
   
       print FILE "<tr class='normal'>\n";
       print FILE "<td class='cell1' align='left'><font class='white'><b>Most connects to the server</b></font></td>\n";
       print FILE "<td class='cell2' align='left'>" . $high->{'high'} . " connects</td>\n";
       print FILE "<td class='cell3' align='right'><font class='contrast'>" .
               $stats->{'players'}{$high->{'wonId'}}->{'name_html'} . "</font></td>\n";
       print FILE "<td class='cell3' align='center'>[ ";
       print FILE "<a href='players/$high->{'wonId'}.html'>stats</a> ]</td>\n</tr>\n";
    }
    print FILE tEnd();

    if($config{'game_type'} eq "cstrike")
    {
        print FILE "<br>", tStart();
        print FILE "<tr class='headerbg'><td><font class='header'>Awards for counter strike</font></td></tr>\n";
        print FILE tEnd();
        
        $high = getCstrikeAward("bPlants");
        
        if($high->{'high'} > 0)
        {
            print FILE tStart();
            print FILE "<tr class='normal'>\n";
            print FILE "<td class='cell1' align='left' width='363'><font class='white'><b>Most bombs planted</b></font></td>\n";
            print FILE "<td class='cell2' align='left' width='114'>" . $high->{'high'} . " plantings</td>\n";
            print FILE "<td class='cell3' align='right' width='114'><font class='contrast'>" .
                    $stats->{'players'}{$high->{'wonId'}}->{'name_html'} . "</font></td>\n";
            print FILE "<td class='cell3' align='center' width='114'>[ ";
            print FILE "<a href='players/$high->{'wonId'}.html'>stats</a> ]</td>\n</tr>\n";
        }
    
        $high = getCstrikeAward("bDets");
    
        if($high->{'high'} > 0)
        {
            print FILE "<tr class='normal'>\n";
            print FILE "<td class='cell1' align='left'><font class='white'><b>Most bombs detonated</b></font></td>\n";
            print FILE "<td class='cell2' align='left'>" . $high->{'high'} . " explosions</td>\n";
            print FILE "<td class='cell3' align='right'><font class='contrast'>" .
                    $stats->{'players'}{$high->{'wonId'}}->{'name_html'} . "</font></td>\n";
            print FILE "<td class='cell3' align='center'>[ ";
            print FILE "<a href='players/$high->{'wonId'}.html'>stats</a> ]</td>\n</tr>\n";
        }
    
        $high = getCstrikeAward("bDefuse");
    
        if($high->{'high'} > 0)
        {
            print FILE "<tr class='normal'>\n";
            print FILE "<td class='cell1' align='left'><font class='white'><b>Most bombs defused</b></font></td>\n";
            print FILE "<td class='cell2' align='left'>" . $high->{'high'} . " defuses</td>\n";
            print FILE "<td class='cell3' align='right'><font class='contrast'>" .
                    $stats->{'players'}{$high->{'wonId'}}->{'name_html'} . "</font></td>\n";
            print FILE "<td class='cell3' align='center'>[ ";
            print FILE "<a href='players/$high->{'wonId'}.html'>stats</a> ]</td>\n</tr>\n";
        }
    
        $high = getCstrikeAward("hTouch");
    
        if($high->{'high'} > 0)
        {
            print FILE "<tr class='normal'>\n";
            print FILE "<td class='cell1' align='left'><font class='white'><b>Most hostages touched</b></font></td>\n";
            print FILE "<td class='cell2' align='left'>" . $high->{'high'} . " molestations</td>\n";
            print FILE "<td class='cell3' align='right'><font class='contrast'>" .
                    $stats->{'players'}{$high->{'wonId'}}->{'name_html'} . "</font></td>\n";
            print FILE "<td class='cell3' align='center'>[ ";
            print FILE "<a href='players/$high->{'wonId'}.html'>stats</a> ]</td>\n</tr>\n";
        }
    
        $high = getCstrikeAward("hRescue");
    
        if($high->{'high'} > 0)
        {
            print FILE "<tr class='normal'>\n";
            print FILE "<td class='cell1' align='left'><font class='white'><b>Most hostage rescues</b></font></td>\n";
            print FILE "<td class='cell2' align='left'>" . $high->{'high'} . " people saved</td>\n";
            print FILE "<td class='cell3' align='right'><font class='contrast'>" .
                    $stats->{'players'}{$high->{'wonId'}}->{'name_html'} . "</font></td>\n";
            print FILE "<td class='cell3' align='center'>[ ";
            print FILE "<a href='players/$high->{'wonId'}.html'>stats</a> ]</td>\n</tr>\n";
        }
    
        $high = getCstrikeAward("hKills");
    
        if($high->{'high'} > 0)
        {
            print FILE "<tr class='normal'>\n";
            print FILE "<td class='cell1' align='left'><font class='white'><b>Most hostages killed</b></font></td>\n";
            print FILE "<td class='cell2' align='left'>" . $high->{'high'} . " innocent dead</td>\n";
            print FILE "<td class='cell3' align='right'><font class='contrast'>" .
                    $stats->{'players'}{$high->{'wonId'}}->{'name_html'} . "</font></td>\n";
            print FILE "<td class='cell3' align='center'>[ ";
            print FILE "<a href='players/$high->{'wonId'}.html'>stats</a> ]</td>\n</tr>\n";
        }
    
        $high = getCstrikeAward("vKills");
    
        if($high->{'high'} > 0)
        {
            print FILE "<tr class='normal'>\n";
            print FILE "<td class='cell1' align='left'><font class='white'><b>Best VIP assasin</b></font></td>\n";
            print FILE "<td class='cell2' align='left'>" . $high->{'high'} . " VIPs offed</td>\n";
            print FILE "<td class='cell3' align='right'><font class='contrast'>" .
                    $stats->{'players'}{$high->{'wonId'}}->{'name_html'} . "</font></td>\n";
            print FILE "<td class='cell3' align='center'>[ ";
            print FILE "<a href='players/$high->{'wonId'}.html'>stats</a> ]</td>\n</tr>\n";
        }
    
        $high = getCstrikeAward("Tjoins");
    
        if($high->{'high'} > 0)
        {
            print FILE "<tr class='normal'>\n";
            print FILE "<td class='cell1' align='left'><font class='white'><b>Most dedicated terrorist</b></font></td>\n";
            print FILE "<td class='cell2' align='left'>" . $high->{'high'} . " joins</td>\n";
            print FILE "<td class='cell3' align='right'><font class='contrast'>" .
                    $stats->{'players'}{$high->{'wonId'}}->{'name_html'} . "</font></td>\n";
            print FILE "<td class='cell3' align='center'>[ ";
            print FILE "<a href='players/$high->{'wonId'}.html'>stats</a> ]</td>\n</tr>\n";
        }
    
        $high = getCstrikeAward("CTjoins");
    
        if($high->{'high'} > 0)
        {
            print FILE "<tr class='normal'>\n";
            print FILE "<td class='cell1' align='left'><font class='white'><b>Most dedicated counter-terrorist</b></font></td>\n";
            print FILE "<td class='cell2' align='left'>" . $high->{'high'} . " joins</td>\n";
            print FILE "<td class='cell3' align='right'><font class='contrast'>" .
                    $stats->{'players'}{$high->{'wonId'}}->{'name_html'} . "</font></td>\n";
            print FILE "<td class='cell3' align='center'>[ ";
            print FILE "<a href='players/$high->{'wonId'}.html'>stats</a> ]</td>\n</tr>\n";
            
            print FILE tEnd();
        }
    }

    $rank = 1;

    print FILE "<br>", tStart();
    print FILE "<tr class='headerbg'><td><font class='header'>Awards for highest kills with a weapon</font></td></tr>\n";
    print FILE tEnd();

    print FILE tStart();

    foreach $weapon ( sort
              { $stats->{'server'}{'weapons'}{$b}
            <=> $stats->{'server'}{'weapons'}{$a} }
              keys %{$stats->{'server'}{'weapons'}} )
    {
        $high = getWeaponAward($weapon);
#print "-------------------------------\n";    
#print "weapon '$weapon'\n";    
#print "high:high '$high->{'high'}'\n";    
#print "high:wonId '$high->{'wonId'}'\n";    
#print "high:total '$high->{'total'}'\n";    
        print FILE "<tr height='75'>\n";
        print FILE "<td class='cell1' align='center' width='363'>\n";
        if(-e "$config{'abs_img_dir'}/weapons/$config{'game_type'}/$weapon.jpg")
        {
            print FILE "<img src='$config{'rel_img_dir'}/weapons/$config{'game_type'}/$weapon.jpg' border='0'>\n";
        }
        else
        {
            print FILE "<font class='normal'><b>$weapon</b></font>\n";
        }
        print FILE "</td>\n";
        
        print FILE "<td class='cell2' align='center' width='364'>\n";
        
        print FILE "<table width='360' border='0' cellspacing='2' cellpadding='5'>\n";
    
        print FILE "<tr valign='middle' class='normal'>\n";

        if ($stats->{'players'}{$high->{'wonId'}} &&
            $stats->{'players'}{$high->{'wonId'}}{'name_html'} )
        {
            print FILE "<td class='cell3' align='right' width='246'><font class='white'><b>" . $stats->{'players'}{$high->{'wonId'}}{'name_html'} . "</b></font></td>\n";
        }
        else
        {
            print FILE "<td class='cell3' align='right' width='246'><font class='white'><b>" . "Nobody!" . "</b></font></td>\n";
        }

        print FILE "<td class='cell3' align='center' width='88'>[ ";
        print FILE "<a href='players/$high->{'wonId'}.html'>stats</a> ]</td>\n</tr>\n";
        
        print FILE "<tr valign='middle'>\n";
        print FILE "<td class='cell1'><font class='small'>" . getBar("",$high->{'high'}, $high->{'total'}) . "</font></td>\n";
        print FILE "<td class='cell1' align='center'><font class='normal'>" . $high->{'total'} . "</font></td>\n";
        print FILE "</tr></table>\n";
        
        print FILE "</td>\n";
        print FILE "</tr>\n";
        $rank++;
        if($rank % $config{'sleep_interval'} == 0) { sleep $config{'sleep_time'}; }
    }
    print FILE tEnd(), "<br>\n";
    
    print FILE pagefoot();
    close(FILE);    
}


#------------------------------------------------------------------------------
# Get awards specific to Counterstrike
sub getCstrikeAward
{
    my($var,$player,%high,$cur);
    $var = shift;

    $high{'count'} = 0;
    $high{'high'}  = 0;
    $high{'total'} = 0;

    foreach $player(@players)
    {
        if(length($player) >= 1)
        {
            $cur = $stats->{'players'}{$player}{'cstrike'}{$var} || 0;
            
            if( $cur > $high{'high'} && canReportPlayer($player) )
            {
                $high{'high'} = $cur;
                $high{'wonId'} = $player;
            }
            $high{'total'} += $cur;
        }
    }
    if(!$high{'wonId'})
    {
        $high{'high'} = 0;
    }
    return(\%high);
}

#------------------------------------------------------------------------------
# Figure out who was killed the most
sub getSuicideAward
{
    my($weapon,$player,%high,$cur);

    $high{'count'} = 0;
    $high{'high'}  = 0;
    $high{'total'} = 0;

    foreach $player(@players)
    {
        if(length($player) >= 1)
        {
            $cur = $stats->{'players'}{$player}->{'suicides'};

            if($cur > $high{'high'} && canReportPlayer($player))
            {
                $high{'high'} = $cur;
                $high{'wonId'} = $player;
            }
            $high{'total'} += $cur;
        }
    }
    return(\%high);
}

#------------------------------------------------------------------------------
# Figure out who connected the most times
sub getConnectAward
{
    my($weapon,$player,%high,$cur);

    $high{'count'} = 0;
    $high{'high'}  = 0;
    $high{'total'} = 0;

    foreach $player(@players)
    {
        if(length($player) >= 1)
        {
            $cur = $stats->{'players'}{$player}->{'connects'} || 0;

            if($cur > $high{'high'} && canReportPlayer($player))
            {
                $high{'high'} = $cur;
                $high{'wonId'} = $player;
            }
            $high{'total'} += $cur;
        }
    }
    return(\%high);
}

#------------------------------------------------------------------------------
# Figure out who played the most on this server
sub getTimeAward
{
    my($weapon,$player,%high,$cur);

    $high{'count'} = 0;
    $high{'high'}  = 0;
    $high{'total'} = 0;

    foreach $player(@players)
    {
    if(length($player) >= 1)
    {
        $cur = $stats->{'players'}{$player}->{'total_time'};
        
        if($cur > $high{'high'} && canReportPlayer($player))
        {
            $high{'high'} = $cur;
            $high{'wonId'} = $player;
        }
        $high{'total'} += $cur;
    }
    }
    return(\%high);
}

#------------------------------------------------------------------------------
# Gets the weapon award for the specified implement.
#     If no award was found then $high{'wonId'} will = ""
sub getWeaponAward
{
    my($weapon,$player,%high,$cur);
    $weapon = shift;

    $high{'count'} = 0;
    $high{'high'}  = 0;
    $high{'total'} = 0;
    $high{'wonId'} = "";

    foreach $player(@players)
    {
        if(length($player))
        {
            $cur = $stats->{'players'}{$player}->getWeaponUses($weapon);
            if($cur && $cur > $high{'high'} && canReportPlayer($player))
            {
                $high{'high'} = $cur;
                $high{'wonId'} = $player;
                $high{'count'} += 1;
            }
            $high{'total'} += $cur;
        }
    }
    return(\%high);
}

#------------------------------------------------------------------------------
# Figure out what a player's favorite weapon was
sub getFavWeapon
{
    my ($k,$v,$weap,$hiCount,@favWeap);
    $k = shift;$v = shift;
    $hiCount = 0;
    foreach $weap (keys %{$stats->{'players'}{$k}->{'victims'}{$v}{'weapons'}})
    {
        if($stats->{'players'}{$k}->{'victims'}{$v}{'weapons'}{$weap} > $hiCount)
        {
            $favWeap[0] = $weap; 
            $hiCount = $stats->{'players'}{$k}->{'victims'}{$v}{'weapons'}{$weap};
        }
    }
    $favWeap[1] = $hiCount;
    return(@favWeap);
}

#------------------------------------------------------------------------------
# Use Data::Dumper to save all our data structures to disk
sub writeData
{
    # Before we dump the player statistics, lets make sure the target subdir exists
    if(!(-e "data/players/$cfgFile"))
    {
        if($config{'os'} eq "unix")
        {
            mkdir ("data", 0770);
            mkdir ("data/players", 0770);
            mkdir ("data/players/$cfgFile", 0770);
        }
        else
        {
            mkdir ("data", 0770);
            mkdir ("data\\players", 0770);
            mkdir ("data\\players\\$cfgFile", 0770);
        }
    }

    # First dump the $stats->{'old_logs'} structure and free it from memory when done
    open(FILE, ">data/$cfgFile-old_logs-logData.txt")  || die $!;
    print FILE Dumper($stats->{'old_logs'});
    close(FILE);
    undef $stats->{'old_logs'};

    # Then dump the $stats->{'live'} structure and free it from memory when done
    open(FILE, ">data/$cfgFile-live-logData.txt")  || die $!;
    print FILE Dumper($stats->{'live'});
    close(FILE);
    undef $stats->{'live'};

    # Now dump the $stats->{'server'} structure and free it from memory when done
    open(FILE, ">data/$cfgFile-server-logData.txt")  || die $!;
    print FILE Dumper($stats->{'server'});
    close(FILE);
    undef $stats->{'server'};

    # Finally dump each member of the $stats->{'players'} structure to disk
    #    and free each from memory when done
    my $plr;
    foreach $plr (@players)
    {
        open(FILE, ">data/players/$cfgFile/$plr.dat");
        print FILE Dumper($stats->{'players'}{$plr});
        close(FILE);
        
        undef $stats->{'players'}{$plr};
    } 
}

#------------------------------------------------------------------------------
# Get the length of the graphical bar for a specific chart
sub getBar
{
    my ($num,$tot,$pcnt,$img,$text,$flag,$mult,$pathPrefix);
    $pathPrefix = shift;
    $num = shift;if(!$num) { $num = 0;}
    $tot = shift;if(!$tot) { $tot = 0;}
    $flag = shift;if(!$tot) { $flag = 0;}

    if($tot < 1) {return(0);}
    if($flag && $flag eq "no text") { $mult = 2.0; } else { $mult = 1.75; }
    $pcnt = sprintf("%0.1f", ($num/$tot)*100);
    
    if($pcnt >= 50) 
    { 
        $img = "green.jpg"; 
    }
    elsif($pcnt >= 25) 
    { 
        $img = "yellow.jpg"; 
    }
    else 
    { 
        if($pcnt < 1 && $pcnt > 0) 
        { 
            $pcnt = "1.0"; 
        }
        $img = "red.jpg"; 
    }

    if($pcnt > 0)
    {
        $text = "<img src='$config{'rel_img_dir'}/$img' border='0' width='" . $pcnt * $mult . "' height='20'>";
    }
    else
    {
        $text = "<font class='small'>0 (0.0%)</font>\n";
    }
    
    unless($flag && $flag eq "no text")
    {
        if($pcnt > 0)
        {
            $text .= "<font class='small'>$num ($pcnt \%)</font>";
        }
    }
    return($text);
}

#------------------------------------------------------------------------------
# Players often use "illegal" characters in their names, so we must make them
#     legal for use in filenames
sub cleanName
{
    # rename non filesystem safe chars
    my $name = shift;
    if(length($name) < 1) { return($name); }
    $name =~ s/ /_/g;
    $name =~ s/\=/\[EQUAL\]/g;
    $name =~ s/\,/\[CMA\]/g;
    $name =~ s/;/\[SMICLN\]/g;
    $name =~ s/:/\[CLN\]/g;
    $name =~ s/\"/\[DBLQT\]/g;
    $name =~ s/\'/\[SNGLQT\]/g;
    $name =~ s/\{/\[LB2\]/g;
    $name =~ s/\}/\[RB2\]/g;
    $name =~ s/\+/\[PLS\]/g;
    $name =~ s/\-/\[DSH\]/g;
    $name =~ s/\(/\[LPRN\]/g;
    $name =~ s/\)/\[RPRN\]/g;
    $name =~ s/\*/\[STR\]/g;
    $name =~ s/\&/\[AND\]/g;
    $name =~ s/\|/\[PIPE\]/g;
    $name =~ s/\\/\[BSLSH\]/g;
    $name =~ s/\//\[FSLSH\]/g;
    $name =~ s/\?/\[QUESTN\]/g;
    $name =~ s/\</\[LT\]/g;
    $name =~ s/\>/\[GT\]/g;
    $name =~ s/\./\[DOT\]/g;
    $name =~ s/\!/\[XCLM\]/g;
    $name =~ s/\@/\[AT\]/g;
    $name =~ s/\#/\[PND\]/g;
    $name =~ s/\$/\[DLLR\]/g;
    $name =~ s/\%/\[PRCNT\]/g;
    $name =~ s/\^/\[CRT\]/g;
        
    if(length($name) > 100) { $name = substr($name, 0, 100); }

    return($name);
}

#------------------------------------------------------------------------------
# If the HTML help file doesn't already exist, we'll create it for them
sub writeHelpFile
{
    unless(-e "$config{'abs_html_dir'}/help.html")
    {
        my $helpText;
        $helpText = pagehead("Help","") . "<br>\n";
        $helpText .= <<EOM;

<br><br><center>
<table width='750' border='0' cellspacing='3' cellpadding='3'>
<tr class='cell1'><td><A name=top>halfStats</a> FAQ&amp;A:</td></tr>
<tr class='cell3'><td><ol>
<li><a href="#halfStats">What is halfStats?</a></li>
<li><a href="#show">How come I don't show up on the stats?</a></li>
<li><a href="#efficiency">How is efficiency determined?</a></li>
<li><a href="#skill">What's this skill rating thing?</a></li>
<li><a href="#sort">How do I sort by something other than skill rating?</a></li>
<li><a href="#aliases">What does number of aliases used mean?</a></li>
<li><a href="#stats">How come some of these stats don't seem correct?</a></li>
<li><a href="#award">I want an award! How do I do it?</a></li>
<li><a href="#back">How do I get back to the server stats?</a></li>
<li><a href="#download">How do I go about getting this great, superb, wonderful, 
  ingenious program for my server?</a></li>
<li><a href="#question">I have another question!!!</a></li></ol></td></tr>
</table><br><br>

<table width='750' border='0' cellspacing='3' cellpadding='5'>

<tr class='cell1'><td><a NAME=halfStats><b>What</b></a><b> is halfStats?</b></td></tr>
<tr class='cell3'><td><blockquote>halfStats is a log analzyer for half-life servers created by Zach 
  Karpinski. It takes log files generated by a linux, freebsd or windows half-life server and 
  turns them into pretty stats that can be put onto a web server. It works for 
  internet or LAN servers and has a variety of options you can use to customize 
  a statistics page. </blockquote><br><a href="#top">|back to top|</a></td></tr>
<tr bgcolor='black'><td>&nbsp;</td></tr>

<tr class='cell1'><td><a name=show><b>How</b></a><b> come I don't show up on the stats?</b></td></tr>
<tr class='cell3'><td><blockquote>There are a few reasons this could be so:
<ol>
  <li>The stats program may not have been run on this server yet.</li>
  <li>You did not meet the minimum skill rating or number of kills set by the 
    administrator to be listed. </li>
  <li>The server did not pick you up in the logs correctly. <br>|<a href="#top">back 
    to top|</a></li>
</ol></td></tr>

<tr bgcolor='black'><td>&nbsp;</td></tr>

<tr class='cell1'><td><a name=efficiency><b>How</b></a><b> is efficiency determined?</b></td></tr>
<tr class='cell3'><td><blockquote>Efficiency determines how efficient a player is with his kill/death ratio.
player efficiency = player kills/player deaths </blockquote><br><a href="#top">|back to top|</a></td></tr>

<tr bgcolor='black'><td>&nbsp;</td></tr>

<tr class='cell1'><td><a name=skill><b>What</b></a><b> is this skill rating thing?</b></td></tr>
<tr class='cell3'><td><blockquote>Skill is an arbitrary number. It is made up. It has nothing to do with your 
  deaths or your kills. Skill starts out at 1000. It will generally not go below 980. When you kill someone, 
that person's skill gets divided by your skill, then  multiplied by 10 and finally added to that person's current 
skill.<br><br>killer skill = ((victim skill / killer skill) * 10) + killer skill <br><br>
So now you may ask how you lose skill? Basically you lose what your victor gained:<br>
victim (new) skill = victim (old) skill - ((victim skill / killer skill) * 10) 
</blockquote><br><a href="#top">|back to top|</a></td></tr>

<tr bgcolor='black'><td>&nbsp;</td></tr>

<tr class='cell1'><td><b><a name=sort>How</a> do I sort by something other than skill rating?</b></td></tr>
<tr class='cell3'><td><blockquote>On the <a href="skill.html">players</a> page click on a column name you wish 
  to sort by. You can sort by (in ascending order) kills, deaths, efficiency, kills per hour 
  and skill rating. </blockquote><br><a href="#top">|back to top|</a></td></tr>

<tr bgcolor='black'><td>&nbsp;</td></tr>

<tr class='cell1'><td><b><a name=aliases>What</a> does number of aliases used mean?</b></td></tr>
<tr class='cell3'><td><blockquote>The word alias is another word for handle which is another word for codename 
  which is another word for other name which is another word for.. You get the idea.
Basically this function keeps tracks of all the different names this person  uses in game.
It helps admins keep track of unorderly players and allows onlookers to make fun of all your names ;) 
</blockquote><br><a href="#top">|back to top|</a></td></tr>

<tr bgcolor='black'><td>&nbsp;</td></tr>

<tr class='cell1'><td><b><a name=stats>How</a> come some of these stats don't seem correct?</b></td></tr>
<tr class='cell3'><td><blockquote>There are a few reasons this might happen.  First off the server may be logging
stats with one of three different store methods. <br><br>The first method and arguably the
most accurate, is by wonid.  Everytime you play, won.net lets the server know your
unique id.  The problem is, there is more than one won.net server, and they dont 
talk to each other.  At this point, you can have up to three different wonids, and
when you connect, you might be using any one of the three.  If you want to make
sure you only use one won.net authentication server, go into your halflife\\valve
directory and open a file called "woncomm.lst".  Go to the part that says Auth, and
remove all but one of the servers.  There are three, West, Central and East.  Keep the
one closest to your place of residence.<br><br>Another way to store is by IP.  This
way works great for LAN games and for LPB servers as many DSL, EDU and cable connections
dont reset IPs that often.  However, for modemers and some cable users, its not good.
You probably wont see this option much on internet servers. <br><br>The other way is to
store by name, which is the most innaccurate way.  Its a bit more forgiving in terms of
what it will log but your stats arent sticky like they are with wonid and ip.<br><br>
Unfortunately the Half-Life logging system is not perfect, and with less than 
  perfect logs you get less than perfect stats.
Ideally all stats work correctly and there shouldn't be a problem.
However, sometimes when a person joins a game in the middle of log opening 
  or closing (ie server start or level change), the player does not get closed 
  out of logs properly and shows false numbers in some fields ( ie. kills per 
  hour, time on server). Life sucks anyway :) </blockquote><br><a href="#top">|back to top|</a></td></tr>

<tr bgcolor='black'><td>&nbsp;</td></tr>

<tr class='cell1'><td><b><a name=award>I</a> want an award! How do I do it?</b></td></tr>
<tr class='cell3'><td><blockquote>Hit ~ and type "unbind all" then hit enter</blockquote><br><a href="#top">|back to top|</a></td></tr>

<tr bgcolor='black'><td>&nbsp;</td></tr>

<tr class='cell1'><td><b><a name=back>How</a> do I get back to the server stats?</b></td></tr>
<tr class='cell3'><td><blockquote>First, smack yourself... Now that you're done with that, click a link any link 
  at the top that says either : main, awards, players, or server. <br><a href="#top">|back 
  to top|</a></td></tr>

<tr bgcolor='black'><td>&nbsp;</td></tr>

<tr class='cell1'><td><b><a name=download>How</a> do I go about getting this great, superb, wonderful, 
  ingenious program for my server?</b></td></tr>
<tr class='cell3'><td><blockquote>Aww, you think it's that great? Well go to the &quot;all your halfStats needs 
  homepage&quot; <a href="http://www.halfstats.com" target="_blank">http://www.halfstats.com 
  </a>. Tell em Curve sent ya ;) </blockquote><br><a href="#top">|back to top|</a></td></tr>

<tr bgcolor='black'><td>&nbsp;</td></tr>

<tr class='cell1'><td><b><a name=question>I</a> have another question!!!</b></td></tr>
<tr class='cell3'><td><blockquote>Holy hell, there's one in every bunch.
If it's a REAL question about something that after much pondering you still 
  don't understand, feel free to <a href='http://boards.gamers.com/messages/overview.asp?name=halfStats&page=1' target='_top'>check out the halfStats message boards</a> to ask a question. </blockquote><br><a href="#top">|back to top|</a></td></tr>

</table>
<br><br><br></center>

EOM

$helpText .= pagefoot();

        open(FILE, ">$config{'abs_html_dir'}/help.html");
        print FILE $helpText;
        close(FILE);
    }
}

#------------------------------------------------------------------------------
# Calculate the difference between passed-in time and right now
sub calcTime
{
    my $startTime = shift;
    return (time() - $startTime);
    
}

#------------------------------------------------------------------------------
# See if we can report the player based on user settings for minimum skill
#     level and number of kills.  This is used to determine whether to create
#     the player HTML file and links to him.
sub canReportPlayer
{
    my ($plr) = shift;

    # Check for missing data elements and avoid lookup if that's the case!
    if ( !$plr ) { return(0) };
    if ( !$config{'minimum_kills'} ) { return(0) };
    if ( !$stats->{'players'} ) { return(0) };
    if ( !$stats->{'players'}{$plr} ) { return(0) };
    if ( !$stats->{'players'}{$plr}->getKills() ) { return(0) };
    if ( !$stats->{'players'}{$plr}->{'skill'} ) { return(0) };
    if ( !$stats->{'players'}{$plr}->{'total_time'} ) { return(0) };
    if ( !$stats->{'players'}{$plr}->{'wonId'} ) { return(0) };

    # Test for number of kills...
    if ( $stats->{'players'}{$plr}->getKills() < $config{'minimum_kills'} ) { return(0) };

    # Test for skill level...
    if ( $stats->{'players'}{$plr}->{'skill'} < $config{'minimum_skill'} ) { return(0) };

    # Make sure they've played with us long enough
    if (( $stats->{'players'}{$plr}->{'total_time'} * 60 ) < $config{'minimum_minutes_played'}) { return(0) };

    # Player fulfilled all the criteria - lets report them!
    return(1);
}

#------------------------------------------------------------------------------
# Eliminate all of the players who haven't visited recently or who should be
# ignored as requested by the user by using reiterative 'removePlayer' calls.
sub weedOutPlayers
{
    my @weededPlayers;
    my @weedMe;
    my $vict = '';
    my $plr  = '';
    my $weed = '';
    
    # Scan all the players and determine which ones need to be removed
    foreach $plr (@players)
    {
        # Get rid of stale players
        if ( !isRecent($plr) )
        {
            print "\tRemoving stale player  : $plr\n";
            push (@weedMe, $plr);

            # Free up the memory that the player object was using
            undef $stats->{'players'}{$plr};
        }
        # Get rid of ignored players
        elsif ( !isNotIgnored($plr) )
        {
            print "\tRemoving ignored player: $plr\n";
            push (@weedMe, $plr);
            
            # Free up the memory that the player object was using
            undef $stats->{'players'}{$plr};
        }
        # We're keeping this player, so store it in our "good" list
        else
        {
           push (@weededPlayers, $plr);
        }
    }
    
    # Copy the weeded player list back onto our main array
    @players = @weededPlayers;
                                      
    # Now remove all references to a player in the victims array and then
    # undef them from the data structure to free up memory
    print "\tCleaning up victim lists\n";
    foreach $plr (@players)
    {
        my %weededVictims;
        
        # Copy the player's victim hash for manipulation
        foreach $vict (keys %{$stats->{'players'}{$plr}->{'victims'}})
        {
            $weededVictims{$vict} = 1;
        }
        
        # Cycle through each players' victim list looking for weeds and
        # remove them as necessary
        foreach $vict (keys %weededVictims)
        {
            foreach $weed (@weedMe)
            {
                if ($vict eq $weed)
                {
                    delete($stats->{'players'}{$plr}->{'victims'}{$vict});
                    last;
                }
            }
        }
    }
}

#------------------------------------------------------------------------------
# Read in the ignored_players_file to see who we should forget about
sub loadIgnoredPlayers
{
    if ($config{ignored_players_file})
    {
        open(FILE, "<$config{ignored_players_file}") || die "Cannot find ignored players file: $!\n";
        while (<FILE>) {
            chomp;                      # no newline
            s/#.*//;                    # no comments
            s/^\s+//;                   # no leading white
            s/\s+$//;                   # no trailing white
            next unless length;         # anything left?
            push (@ignored,$_);
            print "\tWill be ignored: $_\n";
        }
        close(CONF);
    }
}

#------------------------------------------------------------------------------
# Determine if the player in question is "ignored" as requested by the user in
# the ignored_players_file variable.  We return FALSE if we should ignore them.
sub isNotIgnored
{
    my ($ply) = shift;
    my ($ignored);
    
    foreach $ignored (@ignored)
    {
        # If we get a match then tell call to ignore player
        if ($ply eq "$ignored") { return(0); } 
    }
    # Player is not in the ignore list
    return(1);
}

#------------------------------------------------------------------------------
# See if the player has visited us within the minimum_visit_frequency number
#     of days.  We return true if they have been here recently.
sub isRecent
{           
    my ($plr) = shift;

    # First check for last connection time
    my $lastConnect = $stats->{'players'}{$plr}->getLastConnect();
    if ( $lastConnect > (time() - ($config{'minimum_visit_frequency'} * 86400)) )               
    {
        return(1);  # Player has been here recently!
    }
    else
    {
        return(0);  # Player is stale, or unknown and assumed stale!
    }
}

#------------------------------------------------------------------------------
# Logging standard code courtesy of Simon Garner
# Taken from www.hlstats.org/log

sub parseline
{
   my $logline = shift;
   if ($logline =~ /^"([^"]+)" ([^"]+) "([^"]+)" ([^"]+) "([^"]+)"(.*)$/)
   {
       # matches events 8,9,10
   
       $team = "";
       $player = $1; # parse out name, uid and team later
       $event1 = $2; # event type - "killed", "attacked", etc.
       $noun1 = $3; # victim name/objective code, etc.
       $event2 = $4; # "with", etc.
       $noun2 = $5; # weapon/victim name, etc.
       $properties = $6; # parse out keys and values later
   }
   elsif ($logline =~ /^"([^"]+)" ([^"]+) "([^"]+)"(.*)$/)
   {
       # matches events 1,4,5,6,7,11,14a,14b
   
       $team = "";
       $player = $1;
       $event1 = $2;
       $noun1 = $3; # weapon/team code/objective code, etc.
       $event2 = "";
       $noun2 = "";
       $properties = $4;
   }
   elsif ($logline =~ /^"([^"]+)" ([^\(]+)(.*)$/)
   {
       # matches events 2,3
   
       $team = "";
       $player = $1;
       $event1 = $2;
       $noun1 = "";
       $event2 = "";
       $noun2 = "";
       $properties = $3;
   }
   elsif ($logline =~ /^Team "([^"]+)" ([^"]+) "([^"]+)"(.*)$/)
   {
       # matches events 12,15
   
       $team = $1; # Team code
       $player = 0;
       $event1 = $2;
       $noun1 = $3;
       $event2 = "";
       $noun2 = "";
       $properties = $4;
   }
   elsif ($logline =~ /^([^"]+) "([^"]+)"(.*)$/)
   {
       # matches events 13,19a,19b,21
   
       $team = "";
       $player = 0;
       $event1 = $1;
       $noun1 = $2;
       $event2 = "";
       $noun2 = "";
       $properties = $3;
   }
}

sub checkPlayer
{
   my($name,$pNum,$wonId,$team) = getPlayerInfo($player);

   if($stats->{'live'}{$pNum}) 
   {
      return(0);
   }

   my($ip,$port) = split(":", $noun1);
                            
   my $cleanName = cleanName($name);
   my $wonIdToSave = $wonId;

   if($config{'store_method'} eq "ip")
   {
      $stats->{'live'}{$pNum} = $ip;
      $names{$cleanName} = $ip;
      $wonId = $ip;
   }
   elsif($config{'store_method'} eq "name")
   {
      $stats->{'live'}{$pNum} = $cleanName;
      $names{$cleanName} = $cleanName;
      $wonId = $cleanName;
   }
   else
   {
      $stats->{'live'}{$pNum} = $wonId;
      $names{$cleanName} = $wonId;
   }
   
   if(!$pNum) { print "\t\tNotice: No player number\n"; }
   if(!$wonId) { print "\t\tNotice: No Won ID\n"; }

   if($wonId && !$stats->{'players'}{$wonId}{'exists'})
   {
      $stats->{'players'}{$wonId} = new Player();
      if($config{'game_type'} ne "cstrike")
      {
         undef $stats->{'players'}{$wonId}{'cstrike'};
      }
      $stats->{'players'}{$wonId}->setWonId($wonId);
      $stats->{'players'}{$wonId}->saveWonId($wonIdToSave);
      $stats->{'server'}{'general'}{'unique'} += 1;
   }
}

# array getPlayerInfo (string player)
sub getPlayerInfo
{
    my ($player) = @_;

    if ($player =~ /^(.+)<(\d+)><(\d+)><(.*)>$/)
    {
        return ($1, $2, $3, $4);
    }
    else
    {
        return ("", 0, 0, "");
    }
}

# hash getProperties (string propstring)
sub getProperties
{
    my ($propstring) = @_;
    my %properties;

    while ($propstring =~ s/^\s*\((\S+)(?: "([^"]+)")?\)//)
    {
        if ($2)
        {
            $properties{$1} = $2;
        }
        else
        {
            $properties{$1} = 1; # for boolean properties per Note (4)
        }
    }

    return %properties;
}


###############################################################################

